/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __APPLICATION_H__
#define __APPLICATION_H__

#include "xml.h"
#include "project.h"
#include "clip.h"
#include "folder.h"
#include "projectwindow.h"
#include "animation.h"
#include "appsignal.h"
#include "appsettings.h"
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk_imlib.h>

/// The list of default project settings
extern ProjectSettings projectSettingsPresets[];

/** The main application object. 
 *  This is the class that contains all global variables for the app.
 *  It is also responsible for doing the main window, Imlib startup, etc.
 */
class Application {
  /// The application clipboard
  Clip *clipboard;
  /// Flag: Is the clipboard a MediaClip
  int clipboardMediaClip;

  /// The animation playback object
  Animation *animation;
  /// The current project window
  ProjectWindow *mainWindow;
  /// The list of clip users to be notifyed whenever a clip changes
  GSList *listeners;
  
  /// The window has been closed, or the user has select Quit. 
  static void appExit (...);
  /// Build & show the splash screen
  void ShowSplashScreen ();
  /// The splash screen widget
  GdkWindow *splashScreen;
  /// pointer to the currently selected project
  Project *project; 

  /// The application settings (colors, CODEC settings, etc.)
  AppSettings *prefs;

public:
  /** Play a clip.
   *  This function will make a new Animation object and tell it to play the clip */
  void PlayClip (MediaClip *clip, int startFrame, int endFrame);

  /// Constructor - call from main.
  Application (int argc, char **argv);


  /// Theese functions gets called from the signal handlers
  /// Make a new project
  void NewProject ();
  /** Close the current project.
   *  @return TRUE if done, FALSE if cancelled by user. */
  int CloseProject ();
  /// open an existing project
  void OpenProject ();
  /** save the current project
   *  @return TRUE if done, otherwise FALSE. */
  int  SaveProject ();
  /** save the current project under a new filename
   *  @return TRUE if done, otherwise FALSE. */
  int SaveProjectAs();
  
  /// Make a new folder under the current one
  void MakeNewFolder (char *name);
  /// Rename the current folder
  int  RenameCurrentFolder(char *newName);
  /// Erase the current folder (and all media within it).
  void EraseCurrentFolder();

  /// Edit the current clip (this should go)
  void EditClip ();

  /// Import a single image as a clip.
  void ImportSingleImage (char *fileName);
  /// Add a single-color clip to the current folder
  int  AddColorClip(char *name);

  /// Emit a 'clip changed' signal
  inline void ClipChanged (MediaClip *clip);
  /** Inform all registered listeners that an event has happened */
  void EmitSignal (AppSignal *signal);
  /// Add a signal listener to the list.
  void RegisterListener (SignalListener *user);
  /// Remove a signal listener from the list
  void DeRegisterListener (SignalListener *user);

  //Misc helper functions:
  /// Get a pointer to the project window object
  inline ProjectWindow *GetProjectWindow ();
  /// Get a pointer to the current project
  inline Project* GetProject ();
  /// Get a pointer to the application preferences
  inline AppSettings *GetPrefs ();  
  /// Get a ptr to the main animation obj.
  inline Animation *GetAnimation ();

  /// Query if clipboard is a media clip.
  int IsClipboardMediaClip ();

  /// Set a new mediaclip into the clipboard.
  void SetClipboard (MediaClip *src);
  /// Set a new clip into the clipboard.
  void SetClipboard (Clip *src);
  /// Get the Clipboard object if it is a mediaclip.
  MediaClip *GetClipboardMediaClip ();
  /// Get the clipboard object if it is a generic clip.
  Clip *GetClipboard ();
  /** Get the clipboard description string
   *  This function returns the result of the clipboard object's GetTypeName */
  char *GetClipboardDesc ();

  /// Build a clip from a token
  Clip *BuildClip (XmlToken *src);
};

inline AppSettings *Application::GetPrefs () {
  return prefs;
}

inline ProjectWindow *Application::GetProjectWindow () {
  return mainWindow;
}

inline Project* Application::GetProject () {
  return project;
}

inline Animation* Application::GetAnimation () {
  return animation;
}
// The global editor window defined in mainwind.cc
extern Application *app;

inline void Application::ClipChanged (MediaClip *clip) {
  AppSignal sig (CLIP_CHANGED, (void *)clip);
  EmitSignal (&sig);
}
  
#endif    //  __APPLICATION_H__






