/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __CLIPVIEW_H__
#define __CLIPVIEW_H__


#include "mclip.h"
#include "framebuffer.h"
#include "cliptimer.h"
#include <gdk/gdk.h>
#include <gdk_imlib.h>
#include <gtk/gtk.h>
#include "gtkanimation.h"

/** The different set of buttons contained in a clip viewer.
 *  This is a bit field */
enum ClipViewerStyle {
  /// Just the buttons for showing clips.
  SIMPLE = 0,
  /// Also the buttons for setting the IN & out points.
  FULL = 1,
  /// Enable timer.
  TIMER = 2,
  /// Make the buttons flat
  FLAT_BUTTONS = 4,
};

class PlayButtonsCallback;

/** A class used for making a GTK clip viewer (with controls & timeline).
 *  This class handles playback, etc. */
class ClipViewer : public SignalListener {
  /// Callback jump table for control buttons
  static PlayButtonsCallback cb;

  /// Pointer to the source clip.
  MediaClip *src;

  /// The animation player object
  Animation *anim;

  /// The widget to show the current frame in
  GtkWidget *clipWidget;

  /// The widget containing the timeline
  GtkWidget *timeLineWidget;
  /// The clip timer object.
  ClipTimer *clipTimer;

  /// The toolbar widget
  GtkWidget *toolbar;
  
  /** Make the main clip player widget & insert it in parent
   *  @param parent the parent widget of the clip
   *  @param w the width of the clip window
   *  @param h the height of the clip window */
  void MakeWidget (GtkWidget *parent, int w, int h, ClipViewerStyle buttons);

 public:
  /// Make a new clipviewer, showing 'clip' in 'parent'.
  ClipViewer (MediaClip *srcClip, GtkWidget *parent, int width, int height, 
              ClipViewerStyle buttons);
  /** Make a new clipviewer, showing 'clip' in dest.
   *  This version of a clip viewer is used by timelines. It doesn't create a 
   *  widget to show the clip in, but relies on dest to do it */
  ClipViewer (MediaClip *srcClip, GtkWidget *parent, Animation *dest,
              ClipViewerStyle buttons);
  
  /** Set a new position in the clip.
   *  This function also updates the show window of the clip */
  void SetPosition (int newPosition);

  /** Get the current frame of the clip */
  inline int GetPosition ();

  /** Set a new source clip for this clip viewer */
  void SetSource (MediaClip *newSource);

  /// Handle a clipchanged signal
  virtual void Signal (AppSignal *signal);

  /** Rerender the clip viewer window.
   *  This function is called whenever the clip displayed in the clip viewer has had
   *  It's data changed. It rerenders the clip & displays it. */
  void Invalidate ();
 private:
  static void _GoStart (GtkWidget *widget, ClipViewer *cv);
  static void _GoBack (GtkWidget *widget, ClipViewer *cv);
  static void _Stop (GtkWidget *widget, ClipViewer *cv);
  static void _Play (GtkWidget *widget, ClipViewer *cv);
  static void _PlayInOut (GtkWidget *widget, ClipViewer *cv);
  static void _Pause (GtkWidget *widget, ClipViewer *cv);
  static void _GoForward (GtkWidget *widget, ClipViewer *cv);
  static void _GoEnd (GtkWidget *widget, ClipViewer *cv);
  static void _SetIn (GtkWidget *widget, ClipViewer *cv);
  static void _SetOut (GtkWidget *widget, ClipViewer *cv);
  static void _ClearInOut (GtkWidget *widget, ClipViewer *cv);
  static void _Timeline_configure (GnomeCanvas *widget, GdkEventExpose *event, ClipViewer *cv);
  /// Object specific widget callbacks
  /// Go 1 frame forward.
  void GoForward ();
  /// Go 1 frame backwards.
  void GoBack ();
  /// Go to IN point (or start frame)
  void GoStart ();
  /// Go to OUT point (or end frame)
  void GoEnd ();
  /// Start playback
  void Play ();
  /// Start playback from IN to OUT points
  void PlayInOut ();
  /// Stop playback
  void Stop ();
  /// Pause playback
  void Pause ();
  /// Set the IN point.
  void SetIn ();
  /// Set the IN point.
  void SetOut ();
  /// Set the IN point.
  void ClearInOut ();
  /// Set the timeline size
  void ConfigureTimeline ();
};

#endif // __CLIPVIEW_H__





