#ifndef __CODEC_H__
#define __CODEC_H__

#include "framebuffer.h"
#include <gtk/gtk.h>
/** Structure used by the CODEC to report info on a given frame. (see Codec::GetFrameInfo).
 *  This structure will be freed by the application and NOT the CODEC */
struct CodecFrameInfo {
  /// The file position of a given frame (not counting the header)
  unsigned int position;
  /// The size of a given frame (in bytes)
  unsigned int size;
};

/** This struct contains all the pointers for all functions a CODEC should use.
 *  At startup, the program looks for a function called GetCodecInfo (). This function
 *  should return a pointer to a Codec structure that will serve as the main entry point
 *  for any codec. */
struct Codec {
  /** Compress a data stream.
   *  This function is called when the codec should compress a data sream. All info 
   *  Should be read from the info structure. The fb pointer in the codec structure
   *  should point to a valid framebuffer. The data is compressed into the data member.
   *  The streamData parameter points to the memory returned by OpenStream.
   *  @param streamData the same as returned by openstream.
   *  @return the number of bytes used to compress the data. */
  int (*Compress) (void *streamData);

  /** Decompress a data stream.
   *  This function takes the data stored in the data member (which has been loaded by the
   *  application) and uncompresses it into the fb member (which has also been allocated by
   *  the application). 
   *  @param streamData the same as returned by openstream. */
  void (*Decompress) (void *streamData);

  /** Show this CODECs settings. 
   *  This function pointer can be NULL if the CODEC doesn't
   *  have any settings. If the pointer is not NULL, the application will  
   *  include the widget into its configurationn panel
   *  @param streamData the context of this codec, NULL for general settings
   *  @param parent the parent for this widget
   *  @return the configuration widget. */
  GtkWidget *(*GetSettingsWidget) (void *streamData, GtkWidget *parent);

  /** Sets a property of this codec
   *  @param streamData the context of this codec, NULL for general preferences
   *  @param propertyName the name of the property to set. 
   *  @param value the value of the property to set. 
   *  @return true on succes, false otherwise  */
  int (*SetProperty) (void *streamData, char *propertyName, char *value);
  
  /** Gets a property of this codec
   *  @param streamData the context of this codec, NULL for general preferences
   *  @param propertyName the name of the property to get.  
   *  @return a string with the propertyvalue  */
  char *(*GetProperty) (void *streamData, char *propertyName);

  /** Gets the current configuration of the codec
   *  @param streamData the context of this codec, NULL for general preferences 
   *  @return an array of string/value pairs */
  char **(*GetProperties) (void *streamData);

  /** Can this CODEC handle its own stream
   *  This functions return true if the CODEC is capable of managing its own stream. 
   *  When this function returns false, the application is to store and read the
   *  compressed data from a stream. 
   *  @return true or false  */
  int (*HasNativeStream) (void);

  /** Open a new stream.
   *  This function is called by the application when the CODEC should prepare for a new
   *  Stream. This function should perform any initalization that is neccessary to handle
   *  (de)compression. 
   *  @param fileName the file name of the stream's data.
   *  @return a pointer to some data that is sent in all calls to Compress & decompress. 
   *          NULL on error or the codec does not support native streams. */
  void* (*OpenStream) (char *fileName);

  /** Close a stream.
   *  This function is called when the application is finished with the current stream.
   *  The CODEC should free the streamData & perform any cleanups neccessary.*/
  void (*CloseStream) (void *streamData);

  /** Get info on a new Frame 
   *  Get info on how much space to allocate before a compression can commence.
   *  This function should return the maximum space needed for the compression. The
   *  actual space used is returned by Compress()
   *  The values are read from the fb data member
   *  @return the amount of data needed (worst case) for compression */
  int (*GetCompressionBufferSize) (int width, int height);
  
  /** Prepare a new Compression stream. 
   *  This function should set up a CODEC to do some compression. The format & size of the
   *  compressed image should be read from the fb member. The codec should make the header
   *  in the memory pointed to by data.
   *  @return the streamData to be used for calls to DoDecompress */
  void* (*PrepareCompressionStream) (void);

  /** Prepare a new Decompression stream. 
   *  This function should set up a CODEC to do some decompression. The format & size of the
   *  decompressed image should be read from the fb member. The codec should make the header
   *  in the memory pointed to by data.
   *  @return the streamData to be used for calls to DoCompress */
  void* (*PrepareDecompressionStream) (void);

  /** Prepare for showing a frame. 
   *  This function is called before a frame is shown, so the application can know what 
   *  data to send to the DeCompressroutine. 
   *  @param streamData the same as returned by openstream.
   *  @param frameNo the frame number to get info for.   */
  CodecFrameInfo* (*GetFrameInfo) (void *streamData, int frameNo);

  /** Tell the CODEC to show its settings.
   *  If this function is NULL, the CODEC has no settings screen. Otherwise, this should
   *  point to a function making a GTK window. */
  void* (*ShowSettings) (void);

  char id[8];
  /** The name of this Codec.
   *  This should be a one-line description (say max 40 chars (or thereabouts)). */
  char *name;
  /// A longer textual description. This is typically a 4-6 line description of the codec.
  char *description;
  /** A pointer to the framebuffer that the codec should use.
   *  When the codec compresses, it should read from this.
   *  When it decompresses, it should write to this. */
  FrameBuffer *fb;
  /** The CODEC compressed data.
   *  When the CODEC compresses, this is where it should dump its compressed data.
   *  When the CODEC decompresses, this is where it should read it's data from.
   *  It is the application's job to allocate the memory for the data area */
  unsigned char *data;
  /// The size of the CODEC's header.
  int headerSize;
};

/** Initialize a CODEC.
 *  This function is called in the plugin. The CODEC shoud fill a Codec structure & 
 *  return a pointer to itself */
Codec *GetCodecInfo ();

#endif


