#include <stdio.h>
#include <gtk/gtk.h>
#include <gtk/gtkenums.h>
#include <gnome.h>
#include "codec.h"
#include "framebuffer.h"
#include "RTjpegcodec.h"
extern "C" {
#include "RTjpeg.h"
}

// Some strings for the info

static char *id = "RTJPG01";
static char *name = "Real Time JPEG";
static char *desc = "Fast JPEG Compression codec, by Justin Schoeman.\n"
                    "This CODEC can encode/decode at high frame rates\n"
                    "for low to medium resolutions. Can also do Motion compression.";

// this is the info we can configure
struct Properties {
  int motionCompress;
  int gopSize;
  int doGOP;
  int mValue;
  int qValue;
  int quality;
};

// the info we need at compression time
struct CodecHeader {
  int width;
  int height;
  unsigned char *tempBuffer;
  unsigned char *tempBuffer2;
  FrameBuffer   *tempFBYUV422;
  FrameBuffer   *tempFBRGB24;
  CodecFrameInfo *codecFrameInfo;
  int currentFrameNr;
  unsigned long tables[128];
  Properties prop;
};

// this contains the general properties for this codec.
static Properties properties;

static Codec codec;
static GtkWidget *settingsWidget;
static GtkWidget *table2;

/*
 * compress the framebuffer data
 */
int RTJPG01DoCompress (void *streamData) {

  CodecHeader *header = (CodecHeader *)streamData;
  Properties *prop = &header->prop;
  int size;

  // Do some basic sanity checking.
  if (codec.fb->GetWidth() != header->width ||
      codec.fb->GetHeight() != header->height) {
    printf ("ERROR: RTjpeg: Sizes do not match\n");
    return 0;
  }

  // Get location to fill
  unsigned char *data = codec.data;

  // make the sourcebuffer usefull for RTjpeg, it needs a YUV422 buffer
  codec.fb->ConvertInto(header->tempFBYUV422);

  // if we need motioncompress..
  if (prop->motionCompress) {
    size = RTjpeg_mcompress((signed char*)data, (unsigned char*)header->tempFBYUV422->GetVideoData(), prop->mValue, prop->qValue);
  }

  // see it we have reached the gop size to fully encode a frame
  if (!prop->motionCompress || (prop->doGOP && header->currentFrameNr++%prop->gopSize == prop->gopSize-1)) {
    size = RTjpeg_compress((signed char*)data, (unsigned char*)header->tempFBYUV422->GetVideoData());
  }

  // return how much memory was used
  return size;
}


/*
 * decompresses the data into the framebuffer
 */
void RTJPG01DoDecompress (void *streamData) {

  CodecHeader *header = (CodecHeader *)streamData;

  // Get location to fetch data from
  unsigned char *data = codec.data;

  RTjpeg_decompress((signed char *)data, header->tempFBYUV422->GetVideoData());
  
  // use the fastest conversion we have.
  RTjpeg_yuvrgb(header->tempFBYUV422->GetVideoData(), header->tempFBRGB24->GetVideoData());

  header->tempFBRGB24->ConvertInto(codec.fb);
  
}

/** 
 * read the data from the header to start decompression
 */
void *RTJPG01DoPrepareDecompressionStream () {

  CodecHeader *header = new CodecHeader ();

  int *data = (int *)codec.data;

  header->width = data[0];
  header->height = data[1];
  for (int i=0; i<128; i++) {
    header->tables[i] = data[2+i];
  }

  RTjpeg_init_decompress(header->tables, header->width, header->height);

  header->codecFrameInfo = new CodecFrameInfo;

  header->tempBuffer = new unsigned char[2*header->width*header->height+(header->width*header->height)/2];
  header->tempFBYUV422 = new FrameBuffer(header->width, header->height, YUV422);
  header->tempFBYUV422->SetVideoData(header->tempBuffer);
  header->tempBuffer2 = new unsigned char[3*header->width*header->height];
  header->tempFBRGB24 = new FrameBuffer(header->width, header->height, RGB24);
  header->tempFBRGB24->SetVideoData(header->tempBuffer2);
  header->currentFrameNr=0;
  return header;
}

/* Prepare a stream for reading. 
   This function opens the file, reads the header, closes the stream & returns the 
   header info */
void *RTJPG01DoOpenStream (char *fileName) {

  // Attempt to open file
  FILE *in = fopen (fileName, "rb");
  if (!in) {
    printf ("ERROR: %s :Unable to open file: %s", id, fileName);
    return NULL;
  }
  
  // Make a new header. This is being sent to all Compress & decompress calls
  CodecHeader *header = new CodecHeader ();
  
  // Get the header data
  fread (&header->width, sizeof (int), 1, in);
  fread (&header->height, sizeof (int), 1, in);
  for (int i=0; i<128; i++) {
    fread (&header->tables[i], sizeof (int), 1, in);
  }

  RTjpeg_init_decompress(header->tables, header->width, header->height);

  header->codecFrameInfo = new CodecFrameInfo;

  header->tempBuffer = new unsigned char[2*header->width*header->height+(header->width*header->height)/2];
  header->tempFBYUV422 = new FrameBuffer(header->width, header->height, YUV422);
  header->tempFBYUV422->SetVideoData(header->tempBuffer);
  header->tempBuffer2 = new unsigned char[3*header->width*header->height];
  header->tempFBRGB24 = new FrameBuffer(header->width, header->height, RGB24);
  header->tempFBRGB24->SetVideoData(header->tempBuffer2);

  // close the file
  fclose (in);

  // Return the header.
  return header;
}

/* Perform cleanups.
   This func. just deletes the header object allocated in DoOpenStream */
void RTJPG01DoCloseStream (void *streamData) {
  CodecHeader *header = (CodecHeader *)streamData;

  delete header->tempBuffer;
  delete header->tempBuffer2;
  delete header->tempFBYUV422;
  delete header->tempFBRGB24;
  delete header->codecFrameInfo;
  delete header;
}

/* Report on a worst-case amount of data needed.
   return the maximum size of the YUV422 buffer */
int RTJPG01DoCompressionBufferSize (int width, int height) {
  return 2*width*height+(width*height)/2;
}

/* Prepare for compression.
   This function builds the header into the data area. */
void *RTJPG01DoPrepareCompressionStream () {
  int *data = (int *)codec.data;
  data[0] = codec.fb->GetWidth ();
  data[1] = codec.fb->GetHeight ();

  // Make a new header. This is being sent to all Compress & decompress calls
  CodecHeader *header = new CodecHeader ();
  header->width = codec.fb->GetWidth ();
  header->height = codec.fb->GetHeight ();
  header->codecFrameInfo = new CodecFrameInfo;

  header->tempBuffer = new unsigned char[2*header->width*header->height+(header->width*header->height)/2];
  header->tempFBYUV422 = new FrameBuffer(header->width, header->height, YUV422);
  header->tempFBYUV422->SetVideoData(header->tempBuffer);
  header->tempBuffer2 = new unsigned char[3*header->width*header->height];
  header->tempFBRGB24 = new FrameBuffer(header->width, header->height, RGB24);
  header->tempFBRGB24->SetVideoData(header->tempBuffer2);
  memcpy(&header->prop, &properties, sizeof(Properties));

  RTjpeg_init_compress(header->tables, header->width, header->height, header->prop.quality);
  RTjpeg_init_mcompress();

  for (int i=0; i<128; i++) {
    data[2+i] = header->tables[i];
  }
  header->currentFrameNr=0;

  // TODO let the compiler calculate this for us.
  codec.headerSize = 130*4;
  return header;
}

/* Calculate the size & position of a framei, RTjpeg is not yet able to do this
 * it currently just returns some dummy values. */
CodecFrameInfo* RTJPG01DoGetFrameInfo (void *streamData, int frameNo) {
  CodecHeader *header = (CodecHeader *)streamData;

  header->codecFrameInfo->size = 100000;
  header->codecFrameInfo->position = 0;

  return header->codecFrameInfo;
}

void RTJPG01DoMCToggleButton(GtkCheckButton *widget, gpointer data) 
{
  gboolean act = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

  properties.motionCompress = act;
  gtk_widget_set_sensitive (table2, act);
}

void RTJPG01DoSpinButton(GtkSpinButton *widget, gpointer data) 
{
  int val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));

  *((int *)data) = val;
}

void RTJPG01DoQualityHScale(GtkAdjustment *adj, gpointer data) 
{
  *((int *)data) = (int)adj->value;
}

GtkWidget* RTJPG01DoGetSettingsWidget (void *streamData, GtkWidget *parent) {

  GtkWidget *frame1;
  GtkWidget *vbox3;
  GtkWidget *hbox1;
  GtkWidget *label6;
  GtkWidget *hscale1;
  GtkWidget *frame2;
  GtkWidget *vbox4;
  GtkWidget *checkbutton1;
  GtkWidget *label7;
  GtkObject *spinbutton1_adj;
  GtkWidget *spinbutton1;
  GtkObject *spinbutton2_adj;
  GtkWidget *spinbutton2;
  GtkWidget *label9;
  GtkWidget *label8;
  GtkObject *spinbutton3_adj;
  GtkWidget *spinbutton3;
  GtkObject *quality_adj;

  CodecHeader *header = (CodecHeader *)streamData;
  Properties *prop;

  if (header == NULL) prop = &properties;
  else prop = &(header->prop);
  
  frame1 = gtk_frame_new (_("RTJpeg Settings"));
  gtk_widget_ref (frame1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "frame1", frame1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (frame1);
  gtk_container_add (GTK_CONTAINER (parent), frame1);

  vbox3 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox3);
  gtk_object_set_data_full (GTK_OBJECT (parent), "vbox3", vbox3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox3);
  gtk_container_add (GTK_CONTAINER (frame1), vbox3);

  hbox1 = gtk_hbox_new (FALSE, 5);
  gtk_widget_ref (hbox1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "hbox1", hbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox1);
  gtk_box_pack_start (GTK_BOX (vbox3), hbox1, FALSE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox1), 8);

  label6 = gtk_label_new (_("Quality :"));
  gtk_widget_ref (label6);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label6", label6,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label6);
  gtk_box_pack_start (GTK_BOX (hbox1), label6, FALSE, FALSE, 0);

  quality_adj = gtk_adjustment_new (prop->quality, 1, 265, 1, 10, 10);
  //gtk_range_set_update_policy(GTK_RANGE(quality_adj), GTK_UPDATE_DISCONTINUOUS);
  
  hscale1 = gtk_hscale_new (GTK_ADJUSTMENT (quality_adj));
  gtk_widget_ref (hscale1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "hscale1", hscale1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hscale1);
  gtk_box_pack_start (GTK_BOX (hbox1), hscale1, TRUE, TRUE, 3);
  gtk_scale_set_value_pos (GTK_SCALE (hscale1), GTK_POS_LEFT);
  gtk_scale_set_digits (GTK_SCALE (hscale1), 0);

  gtk_signal_connect (GTK_OBJECT (quality_adj), "value_changed",
                      GTK_SIGNAL_FUNC (RTJPG01DoQualityHScale),
                      &(prop->quality));

  frame2 = gtk_frame_new (_("Motion compression"));
  gtk_widget_ref (frame2);
  gtk_object_set_data_full (GTK_OBJECT (parent), "frame2", frame2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (frame2);
  gtk_box_pack_start (GTK_BOX (vbox3), frame2, TRUE, TRUE, 0);

  vbox4 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox4);
  gtk_object_set_data_full (GTK_OBJECT (parent), "vbox4", vbox4,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox4);
  gtk_container_add (GTK_CONTAINER (frame2), vbox4);

  checkbutton1 = gtk_check_button_new_with_label (_("Enabled"));
  gtk_widget_ref (checkbutton1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "checkbutton1", checkbutton1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (checkbutton1);
  gtk_box_pack_start (GTK_BOX (vbox4), checkbutton1, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (checkbutton1), 5);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton1), prop->motionCompress);

  gtk_signal_connect (GTK_OBJECT (checkbutton1), "toggled",
                      GTK_SIGNAL_FUNC (RTJPG01DoMCToggleButton),
                      NULL);

  table2 = gtk_table_new (3, 2, FALSE);
  gtk_widget_ref (table2);
  gtk_object_set_data_full (GTK_OBJECT (parent), "table2", table2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (table2);
  gtk_box_pack_start (GTK_BOX (vbox4), table2, TRUE, TRUE, 0);
  gtk_widget_set_sensitive (table2, prop->motionCompress);

  label7 = gtk_label_new (_("Motion Estimation :"));
  gtk_widget_ref (label7);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label7", label7,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label7);
  gtk_table_attach (GTK_TABLE (table2), label7, 0, 1, 0, 1,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label7), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_padding (GTK_MISC (label7), 46, 0);

  spinbutton1_adj = gtk_adjustment_new (prop->mValue, 1, 255, 1, 10, 10);
  spinbutton1 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton1_adj), 1, 0);
  gtk_widget_ref (spinbutton1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "spinbutton1", spinbutton1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spinbutton1);
  gtk_table_attach (GTK_TABLE (table2), spinbutton1, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinbutton1), TRUE);

  gtk_signal_connect (GTK_OBJECT (spinbutton1), "changed",
                      GTK_SIGNAL_FUNC (RTJPG01DoSpinButton),
                      &(prop->qValue));

  spinbutton2_adj = gtk_adjustment_new (prop->qValue, 1, 255, 1, 10, 10);
  spinbutton2 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton2_adj), 1, 0);
  gtk_widget_ref (spinbutton2);
  gtk_object_set_data_full (GTK_OBJECT (parent), "spinbutton2", spinbutton2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spinbutton2);
  gtk_table_attach (GTK_TABLE (table2), spinbutton2, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  gtk_signal_connect (GTK_OBJECT (spinbutton2), "changed",
                      GTK_SIGNAL_FUNC (RTJPG01DoSpinButton),
                      &(prop->mValue));

  label9 = gtk_label_new (_("GOP size :"));
  gtk_widget_ref (label9);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label9", label9,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label9);
  gtk_table_attach (GTK_TABLE (table2), label9, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label9), GTK_JUSTIFY_LEFT);

  label8 = gtk_label_new (_("Motion threshold :"));
  gtk_widget_ref (label8);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label8", label8,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label8);
  gtk_table_attach (GTK_TABLE (table2), label8, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label8), GTK_JUSTIFY_RIGHT);

  spinbutton3_adj = gtk_adjustment_new (prop->gopSize, 2, 100, 1, 10, 10);
  spinbutton3 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton3_adj), 1, 0);
  gtk_widget_ref (spinbutton3);
  gtk_object_set_data_full (GTK_OBJECT (parent), "spinbutton3", spinbutton3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spinbutton3);
  gtk_table_attach (GTK_TABLE (table2), spinbutton3, 1, 2, 2, 3,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinbutton3), TRUE);

  gtk_signal_connect (GTK_OBJECT (spinbutton3), "changed",
                      GTK_SIGNAL_FUNC (RTJPG01DoSpinButton),
                      &(prop->gopSize));

  return frame1;
}

/* This is called by the app before anything else in the codec is called. This should
   also be used for any intialization */
Codec *GetCodecInfo () {
  // Initialize the codec structure
  codec.Compress = RTJPG01DoCompress;
  codec.Decompress = RTJPG01DoDecompress;
  codec.OpenStream = RTJPG01DoOpenStream;
  codec.CloseStream = RTJPG01DoCloseStream;
  codec.GetCompressionBufferSize = RTJPG01DoCompressionBufferSize;
  codec.PrepareCompressionStream = RTJPG01DoPrepareCompressionStream;
  codec.PrepareDecompressionStream = RTJPG01DoPrepareDecompressionStream;
  codec.GetFrameInfo = RTJPG01DoGetFrameInfo;

  codec.GetSettingsWidget = RTJPG01DoGetSettingsWidget;
  settingsWidget = NULL;     // We don't have a settings screen
  codec.name = name;             // Set the codec name
  memcpy(codec.id, id, 8);             // Set the codec name
  // fill in the default settings
  properties.motionCompress = 1;
  properties.gopSize = 5;
  properties.doGOP = 1;
  properties.mValue = 1;
  properties.qValue = 2;
  properties.quality = 230;
  codec.description = desc;      // Set the codec description
  codec.headerSize = 130*4;
  return &codec;
};
