#include <stdio.h>
#include <gtk/gtk.h>
#include <gnome.h>
#include "tricodec.h"
#include "triframe.h"

// Some strings for the info

static char *id = "TRC0001";
static char *name = "Uncompressed variable frame size";
static char *desc = "This codec stores frames so that they can be restored in\n"
                    "different resolutions. Smaller target resolutions use\n"
                    "much less bandwidth. Ideal for preview images";

struct CodecHeader {
  TriFrame *triframe;
  FBVideoFormat format;
  int width;
  int height;
  int bytesPerPixel;
  FrameBuffer *tmpFB;
  unsigned char *tempBuffer;
  CodecFrameInfo *codecFrameInfo;
};


static Codec codec;
static GtkWidget *settingsWidget;

/*
 * transform the framebuffer data to the new triformat.
 */
int TRC0001DoCompress (void *streamData) {

  CodecHeader *header = (CodecHeader *)streamData;

  // Do some basic sanity checking.
  if (codec.fb->GetWidth() != ((CodecHeader *)streamData)->width ||
      codec.fb->GetHeight() != ((CodecHeader *)streamData)->height) {
    printf ("ERROR: Sizes do not match\n");
    return 0;
  }
  if (codec.fb->GetVideoFormat() != ((CodecHeader *)streamData)->format) {
    printf ("ERROR: formats do not match\n");
    return 0;
  }

  // Get location to fill
  unsigned char *data = codec.data;

  header->triframe->MakeFrame((unsigned char *)codec.fb->GetVideoData(), (unsigned char *)data);

  // return how much memory was used
  return codec.fb->GetWidth() * codec.fb->GetHeight () * codec.fb->GetPixelSize ();
}


/*
 * restores the tri data into the framebuffer;
 */
void TRC0001DoDecompress (void *streamData) {

  CodecHeader *header = (CodecHeader *)streamData;

  // Get location to fetch data from
  unsigned char *data = codec.data;

  header->triframe->RestoreFrame((unsigned char *)data, (unsigned char *)header->tmpFB->GetVideoData(), 1);

  header->tmpFB->ConvertInto(codec.fb);

}

void *TRC0001DoPrepareDecompressionStream () {

  CodecHeader *header = new CodecHeader ();

  int *data = (int *)codec.data;

  header->width = data[0];
  header->height = data[1];
  header->format = (FBVideoFormat)data[2];
  header->bytesPerPixel = data[3];

  header->tempBuffer = new unsigned char[header->width*header->height*header->bytesPerPixel];
  header->triframe = new TriFrame(header->width, header->height, 80, header->bytesPerPixel);
  header->tmpFB = new FrameBuffer(header->triframe->GetWidth(1), header->triframe->GetHeight(1), header->format);
  header->tmpFB->SetVideoData(header->tempBuffer);
  header->codecFrameInfo = new CodecFrameInfo;
  return header;
}

/* Prepare a stream for reading. 
   This function opens the file, reads the header, closes the stream & returns the 
   header info */
void *TRC0001DoOpenStream (char *fileName) {

  // Attempt to open file
  FILE *in = fopen (fileName, "rb");
  if (!in) {
    printf ("ERROR: %s :Unable to open file: %s", id, fileName);
    return NULL;
  }
  
  // Make a new header. This is being sent to all Compress & decompress calls
  CodecHeader *header = new CodecHeader ();
  
  // Get the header data from the file
  fread (&header->width, sizeof (int), 1, in);
  fread (&header->height, sizeof (int), 1, in);
  fread (&header->format, sizeof (int), 1, in);
  fread (&header->bytesPerPixel, sizeof (int), 1, in);

  header->tempBuffer = new unsigned char[header->width*header->height*header->bytesPerPixel];
  header->triframe = new TriFrame(header->width, header->height, 80, header->bytesPerPixel);
  header->tmpFB = new FrameBuffer(header->triframe->GetWidth(1), header->triframe->GetHeight(1), header->format);
  header->tmpFB->SetVideoData(header->tempBuffer);
  header->codecFrameInfo = new CodecFrameInfo;

  // close the file
  fclose (in);

  // Return the header.
  return header;
}

/* Perform cleanups.
   This func. just deletes the header object allocated in DoOpenStream */
void TRC0001DoCloseStream (void *streamData) {
  CodecHeader *header = (CodecHeader *)streamData;

  delete header->tmpFB;
  delete header->tempBuffer;
  delete header->triframe;
  delete header->codecFrameInfo;
  delete header;
}
/* Report on a worst-case amount of data needed.
   Since this CODEC does no compression , the calculation is quite simple */
int TRC0001DoCompressionBufferSize (int width, int height) {
  return codec.fb->GetWidth() * codec.fb->GetHeight () * codec.fb->GetPixelSize ();
}

/* Prepare for compression.
   This function builds the header into the data area. */
void *TRC0001DoPrepareCompressionStream () {
  int *data = (int *)codec.data;
  data[0] = codec.fb->GetWidth ();
  data[1] = codec.fb->GetHeight ();
  data[2] = codec.fb->GetVideoFormat ();
  data[3] = codec.fb->GetPixelSize ();

  // Make a new header. This is being sent to all Compress & decompress calls
  CodecHeader *header = new CodecHeader ();
  header->width = codec.fb->GetWidth ();
  header->height = codec.fb->GetHeight ();
  header->format = codec.fb->GetVideoFormat ();
  header->bytesPerPixel = codec.fb->GetPixelSize ();

  // we need some extra space here
  header->tempBuffer = new unsigned char[header->width*header->height*header->bytesPerPixel];
  header->triframe = new TriFrame(header->width, header->height, 80, header->bytesPerPixel);
  header->tmpFB = new FrameBuffer(header->triframe->GetWidth(1), header->triframe->GetHeight(1), header->format);
  header->tmpFB->SetVideoData(header->tempBuffer);
  header->codecFrameInfo = new CodecFrameInfo;

  // TODO please let the compiler decide on this...
  codec.headerSize = 16;
  return header;
}

/* Calculate the size & position of a frame */
CodecFrameInfo* TRC0001DoGetFrameInfo (void *streamData, int frameNo) {
  CodecHeader *header = (CodecHeader *)streamData;

  header->codecFrameInfo->size = header->bytesPerPixel * 
               header->width *
               header->height;
  header->codecFrameInfo->position = header->codecFrameInfo->size*frameNo;
  header->codecFrameInfo->size = header->triframe->GetSize(1)*header->bytesPerPixel;
  return header->codecFrameInfo;
}

// there is not much to configure here...
GtkWidget* TRC0001DoGetSettingsWidget (void *streamData, GtkWidget *parent) 
{
  GtkWidget *frame1;

  frame1 = gtk_frame_new (_("Triframe settings"));
  gtk_widget_ref (frame1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "frame1", frame1,
	                             (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (frame1);
  gtk_container_add (GTK_CONTAINER (parent), frame1);

  return frame1;

}

/* This is called by the app before anything else in the codec is called. This should
   also be used for any intialization */
Codec *GetCodecInfo () {
  // Initialize the codec structure
  codec.Compress = TRC0001DoCompress;
  codec.Decompress = TRC0001DoDecompress;
  codec.OpenStream = TRC0001DoOpenStream;
  codec.CloseStream = TRC0001DoCloseStream;
  codec.GetCompressionBufferSize = TRC0001DoCompressionBufferSize;
  codec.PrepareCompressionStream = TRC0001DoPrepareCompressionStream;
  codec.PrepareDecompressionStream = TRC0001DoPrepareDecompressionStream;
  codec.GetFrameInfo = TRC0001DoGetFrameInfo;

  codec.GetSettingsWidget = TRC0001DoGetSettingsWidget;
  settingsWidget = NULL;     // We don't have a settings screen
  codec.name = name;             // Set the codec name
  memcpy(codec.id, id, 8);             // Set the codec name
  codec.description = desc;      // Set the codec description
  codec.headerSize = 16;          // The size of the header
  return &codec;
};
