#include "codeccentral.h"
#include "tristreamcodec.h"

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gnome.h>

// Some strings for the info

static char *id = "TRC0000";
static char *name = "Multichannel Trinity Stream";
static char *desc = "This codec handles multichannel streams. It is\n"
                    "independent of the video and audio codecs used\n"
                    "to compress the data.";

static Codec codec;


struct Properties {
  char width[10];
  char height[10];
  char numFrames[10];
  char videoCodec[9];
  char audioCodec[9];
};

// this contains the general properties for this codec.
static Properties properties;
static GtkWidget *settingsWidget;
static GtkWidget *vidcodecSettings;
static GtkWidget *vbox2;
static GtkWidget *topParent;
						  
// The structure containing all CODEC info. A pointer to this is sent back to the app.
/// structure used for persistant info for a stream.
struct CodecHeader {
  int numberOfChannels;
  long currentFileOffset;
  long currentFrameNr;
  FILE *streamFile;
  Codec *tricodec;
  void *tricodecData;
  int width;
  int height;
  FBVideoFormat format;
  int bytesPerPixel;
  CodecFrameInfo *codecFrameInfo;
  Properties prop;	// the stream specific properties
};

/* Simple compression. 
   This just copies the pixels in the frame buffer to the appropriate codec */
int TRC0000DoCompress (void *streamData) {
  CodecHeader *header = (CodecHeader *)streamData;
  TRI_CHANNEL_HDR chnHdr;

  chnHdr.size = sizeof(TRI_CHANNEL_HDR);
  chnHdr.channelnr = 0;
  header->currentFrameNr++;
  
  header->tricodec->data = codec.data+sizeof(TRI_CHANNEL_HDR);
  header->tricodec->fb = codec.fb;
  chnHdr.channeldata_size = header->tricodec->Compress(header->tricodecData);

  memcpy(codec.data, &chnHdr, sizeof(TRI_CHANNEL_HDR));

  return chnHdr.channeldata_size + sizeof(TRI_CHANNEL_HDR);
}


/* Simple decompression.
   This just copies the data area to the appropriate codec */
void TRC0000DoDecompress (void *streamData) {
  CodecHeader *header = (CodecHeader *)streamData;

  TRI_CHANNEL_HDR chnHdr;

  chnHdr.size = sizeof(TRI_CHANNEL_HDR);
  chnHdr.channelnr = 0;
 
  header->tricodec->data = codec.data+sizeof(TRI_CHANNEL_HDR);
  header->tricodec->fb = codec.fb;
  header->tricodec->Decompress(header->tricodecData);

}
// mechanism to get some properties of this codec, properties are
// codec specific and are not to be used/relied on by any application
char *TRC0000DoGetProperty (void *streamData, char *propertyName) {
   if (!strcmp(propertyName, "numFrames")) {
	  return properties.numFrames;
	}
}

// mechanism to set some properties of this codec.
int TRC0000DoSetProperty (void *streamData, char *propertyName, char *value) {
   if (!strcmp(propertyName, "numFrames")) {
	  strncpy(properties.numFrames, value, sizeof(properties.numFrames));
	}
   return 0;
}

// get all of the current properties.
char **TRC0000DoGetProperties (void *streamData) {
}

// prepare for decompression from the data area
void *TRC0000DoPrepareDecompressionStream () {
}

/* Prepare a stream for reading. 
   This function opens the file, reads the header, closes the stream & returns the 
   header info */
void *TRC0000DoOpenStream (char *fileName) {

  char propertyBuf[128];

  CodecHeader *header = new CodecHeader ();
  // Attempt to open file
  header->streamFile = fopen (fileName, "rb");
  if (!header->streamFile) {
    printf ("ERROR: Unable to open file: %s", fileName);
    return NULL;
  }

  codec.headerSize = 0;
  
  TRI_HDR tri_hdr;
  TRI_CODEC_HDR tri_codec_hdr;

  fread(&tri_hdr, sizeof(tri_hdr), 1, header->streamFile);
  codec.headerSize += sizeof(tri_hdr);

  int numberOfChannels = tri_hdr.channels;
  sprintf(propertyBuf, "%d", tri_hdr.frames);
  TRC0000DoSetProperty(header, "numFrames", propertyBuf);

  for (int i=0; i < numberOfChannels; i++) {
    // read the codec header to know the codec for this channel
    fread(&tri_codec_hdr, sizeof(tri_codec_hdr), 1, header->streamFile);
    codec.headerSize += sizeof(tri_codec_hdr);
    // get the codec..
    Codec *subCodec = CodecCentral::GetCodecWithID(tri_codec_hdr.codecID);

    if (subCodec == NULL) {
      printf ("unknown codec ID :%s\n", tri_codec_hdr.codecID);
    }
    else {
	   header->tricodec = subCodec;
      unsigned char *hdr = new unsigned char[tri_codec_hdr.codech_size];
      fread(hdr, tri_codec_hdr.codech_size, 1, header->streamFile);
      codec.headerSize += tri_codec_hdr.codech_size;
		header->tricodec->data = hdr;
      header->tricodecData = subCodec->PrepareDecompressionStream();
		delete hdr;
    }

  }
  header->codecFrameInfo = new CodecFrameInfo;

  header->currentFileOffset = ftell(header->streamFile);
  header->currentFrameNr = 0;  // we are positioned at the first frame
																					   
  // Return the header.
  return header;
}

/* Perform cleanups.
   This func. just deletes the header object allocated in DoOpenStream */
void TRC0000DoCloseStream (void *streamData) {
  CodecHeader *header = (CodecHeader *)streamData;

  if (header->streamFile) fclose(header->streamFile);
  delete header->codecFrameInfo;
  delete header;
}
/* Report on a worst-case amount of data needed.
   Since this CODEC uses other codecs, we ask them */
int TRC0000DoCompressionBufferSize (int width, int height) {
  return sizeof(TRI_CHANNEL_HDR)+
     CodecCentral::GetCodecWithID(properties.videoCodec)->GetCompressionBufferSize(width, height);
}

/* Prepare for compression.
   This function builds the header into the data area. */
void *TRC0000DoPrepareCompressionStream () {

  CodecHeader *header = new CodecHeader ();
  TRI_HDR tri_hdr;

  tri_hdr.size = sizeof(TRI_HDR);
  tri_hdr.flags = 0;
  tri_hdr.frames = atoi(TRC0000DoGetProperty(NULL,"numFrames"));
  tri_hdr.channels = 1;
  tri_hdr.bufsize = 0;

  unsigned char *data = (unsigned char *)codec.data;
  memcpy(data, &tri_hdr, sizeof(tri_hdr));
  data+= sizeof(tri_hdr);

  if (!(header->tricodec = CodecCentral::GetCodecWithID(properties.videoCodec))) {
    printf("ERROR: subcodec %s not found\n", properties.videoCodec);
	 return NULL;
  }

  TRI_CODEC_HDR tri_codec_hdr;
  tri_codec_hdr.size = sizeof(tri_codec_hdr);
  memcpy(tri_codec_hdr.codecID, header->tricodec->id, 8);
  tri_codec_hdr.flags = 0;
  tri_codec_hdr.priority = 0;
  tri_codec_hdr.codech_size = header->tricodec->headerSize;
  memcpy(data, &tri_codec_hdr, sizeof(tri_codec_hdr));
  data+= sizeof(tri_codec_hdr);

  header->tricodec->data = data;
  header->tricodec->fb = codec.fb;
  header->tricodecData = header->tricodec->PrepareCompressionStream();
  data+=header->tricodec->headerSize;

  // Make a new header. This is being sent to all Compress & decompress calls
  header->width = codec.fb->GetWidth ();
  header->height = codec.fb->GetHeight ();
  header->format = codec.fb->GetVideoFormat ();
  header->bytesPerPixel = codec.fb->GetPixelSize ();
  header->codecFrameInfo = new CodecFrameInfo;
  header->streamFile = NULL;
  codec.headerSize = (int)data-(int)codec.data;
  return header;
}

/* Calculate the size & position of a frame */
CodecFrameInfo* TRC0000DoGetFrameInfo (void *streamData, int frameNo) {
  CodecFrameInfo *temp2;
  CodecHeader *header = (CodecHeader *)streamData;
  TRI_CHANNEL_HDR chnHdr;

  if (header->currentFrameNr > frameNo) {
    //we have gone too far, just start at the beginning
	 header->currentFrameNr = 0;
	 header->currentFileOffset = codec.headerSize;
  }

  //TODO seek out for the right frame, ie do skipping.

  fseek(header->streamFile, header->currentFileOffset, SEEK_SET);

  header->codecFrameInfo->position = header->currentFileOffset;

  fread(&chnHdr, 1, sizeof(TRI_CHANNEL_HDR), header->streamFile);

  header->codecFrameInfo->size = sizeof(TRI_CHANNEL_HDR)+ chnHdr.channeldata_size;
  header->currentFileOffset += header->codecFrameInfo->size;

  header->currentFrameNr++;


  return header->codecFrameInfo;
};

// callback function for the GUI
void TRC0000codecSelected(GtkWidget *widget, Codec *selCodec) {

  gtk_widget_destroy(vidcodecSettings);
  strncpy(properties.videoCodec, selCodec->id, 8);
  vidcodecSettings = CodecCentral::GetCodecWithID(selCodec->id)->GetSettingsWidget(NULL, topParent);

  gtk_box_pack_start (GTK_BOX (vbox2), vidcodecSettings, TRUE, TRUE, 0);
}

// get the configuration pannel of this codec
GtkWidget *TRC0000DoGetSettingsWidget(void *streamData, GtkWidget *parent) {

  GtkWidget *vbox1;
  GtkWidget *hbox1;
  GtkWidget *label1;
  GtkWidget *combo1;
  GList *combo1_items = NULL;
  GtkWidget *notebook1;
  GtkWidget *hbox2;
  GtkWidget *label4;
  GtkWidget *combo2;
  GtkWidget *label2;
  GtkWidget *vbox3;
  GtkWidget *hbox3;
  GtkWidget *label5;
  GtkWidget *combo3;
  GtkWidget *label3;
  GtkWidget *vidcodecSelection;

  Properties *prop;
  CodecHeader *header = (CodecHeader *)streamData;

  topParent = parent;

  if (header == NULL) prop = &properties;
  else prop = &(header->prop);

  settingsWidget = gtk_frame_new (_("Stream settings"));
  gtk_widget_ref (settingsWidget);
  gtk_object_set_data_full (GTK_OBJECT (parent), "settingsWidget", settingsWidget,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (settingsWidget);
  gtk_container_add (GTK_CONTAINER (parent), settingsWidget);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "vbox1", vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (settingsWidget), vbox1);

  hbox1 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "hbox1", hbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox1, FALSE, FALSE, 0);

  notebook1 = gtk_notebook_new ();
  gtk_widget_ref (notebook1);
  gtk_object_set_data_full (GTK_OBJECT (parent), "notebook1", notebook1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (notebook1);
  gtk_box_pack_start (GTK_BOX (vbox1), notebook1, TRUE, TRUE, 0);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox2);
  gtk_object_set_data_full (GTK_OBJECT (parent), "vbox2", vbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox2);
  gtk_container_add (GTK_CONTAINER (notebook1), vbox2);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox2);
  gtk_object_set_data_full (GTK_OBJECT (parent), "hbox2", hbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox2, FALSE, TRUE, 0);

  // ask for the panel with available codecs, current codec should be selected
  vidcodecSelection = CodecCentral::GetVideoCodecSelectionWidget(parent, prop->videoCodec, TRC0000codecSelected);
  gtk_box_pack_start (GTK_BOX (hbox2), vidcodecSelection, TRUE, TRUE, 0);

  // ask the current codec to show its configuration panel
  vidcodecSettings = CodecCentral::GetCodecWithID(prop->videoCodec)->GetSettingsWidget(NULL, parent);
  gtk_box_pack_start (GTK_BOX (vbox2), vidcodecSettings, TRUE, TRUE, 0);

  label2 = gtk_label_new (_("Video"));
  gtk_widget_ref (label2);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label2", label2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label2);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook1), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook1), 0), label2);

  vbox3 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox3);
  gtk_object_set_data_full (GTK_OBJECT (parent), "vbox3", vbox3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox3);
  gtk_container_add (GTK_CONTAINER (notebook1), vbox3);
  gtk_widget_set_sensitive (vbox3, FALSE);

  hbox3 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox3);
  gtk_object_set_data_full (GTK_OBJECT (parent), "hbox3", hbox3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox3);
  gtk_box_pack_start (GTK_BOX (vbox3), hbox3, FALSE, TRUE, 0);

  label5 = gtk_label_new (_("compression CODEC :"));
  gtk_widget_ref (label5);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label5", label5,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label5);
  gtk_box_pack_start (GTK_BOX (hbox3), label5, FALSE, FALSE, 0);

  combo3 = gtk_combo_new ();
  gtk_widget_ref (combo3);
  gtk_object_set_data_full (GTK_OBJECT (parent), "combo3", combo3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (combo3);
  gtk_box_pack_start (GTK_BOX (hbox3), combo3, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (combo3), 4);

  // audio is not done yet TODO
  label3 = gtk_label_new (_("Audio"));
  gtk_widget_ref (label3);
  gtk_object_set_data_full (GTK_OBJECT (parent), "label3", label3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label3);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook1), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook1), 1), label3);
  gtk_widget_set_sensitive (label3, FALSE);

  return settingsWidget;

}

/* This is called by the app before anything else in the codec is called. This should
   also be used for any intialization */
Codec *GetCodecInfo () {
  // Initialize the codec structure
  codec.Compress = TRC0000DoCompress;
  codec.Decompress = TRC0000DoDecompress;
  codec.OpenStream = TRC0000DoOpenStream;
  codec.CloseStream = TRC0000DoCloseStream;
  codec.GetCompressionBufferSize = TRC0000DoCompressionBufferSize;
  codec.GetProperty = TRC0000DoGetProperty;
  codec.SetProperty = TRC0000DoSetProperty;
  codec.GetProperties = TRC0000DoGetProperties;
  codec.PrepareCompressionStream = TRC0000DoPrepareCompressionStream;
  codec.PrepareDecompressionStream = TRC0000DoPrepareDecompressionStream;
  codec.GetFrameInfo = TRC0000DoGetFrameInfo;

  codec.GetSettingsWidget = TRC0000DoGetSettingsWidget;    
  settingsWidget = NULL;             // We don't have a settings screen yet
  codec.name = name;             // Set the codec name
  memcpy(codec.id, id, 8);
  strncpy(properties.videoCodec, "RTJPG01", 8);  //set the default video codec.
  codec.description = desc;      // Set the codec description
  codec.headerSize = 1000;          // The size of the header
  return &codec;
};
