/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __FOLDER_H__
#define __FOLDER_H__


#include <assert.h>
#include "mclip.h"

/** A folder in the project view.
 *  A folder can contain any clips, and other subfolders 
 *  (a folder is also a clip subclass. Unlike any other clips, 
 *  it cannot be inserted in a timeline. It is only used by the project view. */
class Folder : public Clip {
  /// List of clips in this folder.
  GSList *clip;
  /// List of subfolders in this folder.
  GSList *subFolder;
  /// GUI: Is this folder opened in the tree view ?
  int     opened;

  /// The parent folder (NULL if this is the project root folder)
  Folder *parent;

  /** Check if the name of a clip is ok for insertion into a list.
   *  This function makes 2 checks. First, it checks if a clip with the same name exits
   *  in this list. If it does, the function returns 1. Then it goes on to validate if
   *  a file name with the given clip's name could be created. If not, the function 
   *  returns 2. If all is ok, the function returns 0.
   *  @param list the list to check against.
   *  @param clip the clip to check */
  int CheckName (GSList *list, char *clipName);

  
 public:
  Folder ();
  ~Folder ();
  /** Read the contents of this folder from a stream.
   *  This function is called on 'open project' by the project object.
   *  It proceeds to read all clips from the stream (resulting in a hierarchy) .
   *  @param src the XML token to read from. */
  virtual void Load (XmlToken *src);
  /** Write the contents of this folder to a stream.
   *  This function is called on 'save project' by the project object.
   *  It proceeds to write all clips to the stream (resulting in a hierarchy). 
   *  @param dest the destination stream towrite this folder to.*/
  virtual void Save (XmlToken *dest);
  /** Insert a token into this folder. 
   *  This is a helper function used by the Load & Paste functions */
  int Insert (XmlToken *token);
  /** Add a sub-folder to this folder.
   *  This function makes a call to CheckName, and if the name is ok, the folder is added
   *  as a subfolder to this one.
   *  @param folder the folder to add
   *  @return The return code from CheckName */
  int AddSubFolder (Folder *folder);

  /** Make a new (empty) subfolder.
   *  @param fold the folder to add
   *  @return The return code from CheckName */
  int MakeSubFolder (char *name);
  /// return "Folder"
  char *GetTypeName ();
  /** Delete a subfolder from this folder
   *  @param folder the folder to delete. */
  void DeleteSubFolder (Folder *folder);

  /** Add a new clip to this folder
   *  @param newClip the clup to add
   *  @return the return code from CheckName */
  int  AddClip   (MediaClip *newNlip);

  /** Delete a clip from this folder
   *  @param clip the clip to delete */
  void  DeleteClip (MediaClip *clip);

  /// Return the amount of clips in this folder
  inline int GetClipCount ();

  /// Return the n'th clip in this folder
  inline MediaClip *GetClip (int n);

  /// Return the amount of sub folders in this folder
  inline int GetSubFolderCount ();

  /// Return the n'th subfolder
  inline Folder *GetSubFolder (int n);

  /// Return the opened state of the folder
  inline int GetOpened ();

  /// Set the opened state of the folder
  inline void SetOpened (int open);

  /** Check if a given folder name is ok.
   *  @return the return code from CheckName */
  inline int CheckFolderName (char *name);

  /** Check if a given folder name is ok.
   *  @return the return code from CheckName */
  inline int CheckClipName (char *name);

  /// Get the parent of this folder.
  inline Folder *GetParent ();

  /** Erase this folder from the disc.
   *  This function erases a folder, along with any subfolders, subclips, etc. */
  void Erase ();
};

inline int Folder::MakeSubFolder (char *name) {
  Folder *temp = new Folder ();
  temp->SetName (name);
  int status = AddSubFolder (temp);
  if (status)
    delete temp;
  return status;
}

inline int Folder::GetClipCount () {
  return g_slist_length (clip);
}

inline int Folder::GetSubFolderCount () {
  return g_slist_length (subFolder);
}

inline MediaClip *Folder::GetClip (int n) {
  assert (n >= 0 && n < g_slist_length (clip));

  return (MediaClip *)g_slist_nth_data (clip, n);
}

inline Folder *Folder::GetSubFolder (int n) {
  assert (n >= 0 && n < g_slist_length (subFolder));
  return (Folder *)g_slist_nth_data (subFolder, n);
}

inline int Folder::GetOpened () {
  return opened;
}

inline void Folder::SetOpened (int open) {
  opened = open;
}

inline int Folder::CheckFolderName (char *name) {
  return CheckName (subFolder, name);
}

inline int Folder::CheckClipName (char *name) {
  return CheckName (clip, name);
}

inline Folder *Folder::GetParent () {
  return parent;
}
#endif
