/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#include "framebuffer.h"
#include <stdio.h>
#include "assert.h"
#include "color.h"

FrameBuffer::FrameBuffer (int w, int h, FBVideoFormat videoformat, int interlaced = 0, FBAudioFormat audioformat) {
  videoFormat = videoformat;
  switch (videoformat) {
  case RGB32:
  case ARGB32:
    pixelSize = 4;
    break;
  case RGB24:
    pixelSize = 3;
    break;
  case YUV422P:
  case YUV422:
    pixelSize = 2;
    break;
  case YUV853:
  case RGB565:
  case RGB555:
    pixelSize = 2;
    break;
  }

  validVideo = validAudio = 0;

  stride = pixelSize * (interlaced + 1) * w;
  width = w;
  height = h;
  audioFormat = audioformat;
  this->interlaced = interlaced;
  SetField (0);
}

void FrameBuffer::SetPixel (int x, int y, int argb) {
  switch (videoFormat) {
  case RGB32:
  case ARGB32:
    assert (x >= 0 && x < width);
    assert (y >= 0 && y < height);
    *(int*)(GetLine (y) + x * GetPixelSize ()) = argb;
    break;
  case RGB24: {
    argb &= 0xFFFFFF;
    unsigned char *temp = GetLine (y) + x * GetPixelSize ();
    *temp++ = (unsigned char)(argb >> 16);
    *temp++ = (unsigned char)(argb >> 8);
    *temp = (unsigned char)argb;
    break;
    }
  case YUV422: {
    int y2, u, v;
    int r, g, b;
	 r= (argb>>16) & 0xFF;
	 g= (argb>>8) & 0xFF;
	 b= (argb) & 0xFF;

    Color::RGB_to_YUV422Pixel(r, g, b, &y2, &u, &v);

	 unsigned char * data = videoData+(y*GetWidth()*2);

	 *(data+2*x) = y2;
   
	 if (x&1) {
	   *(data+2*x-1) = u;
	   *(data+2*x+1) = v;
	 }

	 }
	 break;
  case YUV422P: {
    int y2, u, v;
    int r, g, b;
	 r= (argb>>16) & 0xFF;
	 g= (argb>>8) & 0xFF;
	 b= (argb) & 0xFF;

    Color::RGB_to_YUV422Pixel(r, g, b, &y2, &u, &v);
	 unsigned char * data = videoData+x/2+(y*GetWidth()/2);
	 *(videoData+x+y*GetWidth()) = y2;
	 *(data+GetWidth()*GetHeight()) = u;
	 *(data+GetWidth()*GetHeight()+(GetWidth()*GetHeight()/2)) = v;
    break;
    }
  default:
    printf ("SetPixel unsupported for this frame buffer format\n");
  }
}

int FrameBuffer::GetPixel (int x, int y) {
  switch (videoFormat) {
  case RGB32:    
    return (*(int*)(GetLine (y) + x * GetPixelSize ())) | 0xFF000000;
  case ARGB32:
    return *(int*)(GetLine (y) + x * GetPixelSize ());
  case RGB24: {
    guchar *temp = GetLine (y) + x * GetPixelSize ();
    int r = *temp++, g = *temp++, b = *temp;
    return ((r << 16) + (g << 8) + b) | 0xFF000000;
  }
  case YUV422: 
    printf("fixme getpixel !! \n");

  case YUV422P: {
    int y2, u, v;
    int r, g, b;
	 unsigned char * data = videoData+x/2+(y*GetWidth()/2);
	 y2 = *(videoData+x+y*GetWidth());
	 u = *(data+GetWidth()*GetHeight());
	 v = *(data+GetWidth()*GetHeight()+(GetWidth()*GetHeight()/2));
    Color::YUV422_to_RGBPixel(y2, u, v, &r, &g, &b);
	 return ((r<<16) + (g << 8) +b) | 0xFF000000;
    break;
    }
  default:
    printf ("GetPixel unsupported for this frame buffer format\n");
    return 0;
  }
}

void FrameBuffer::ConvertInto (FrameBuffer *dest) {
  FrameBuffer *src = this;

  if (dest->GetVideoFormat() != GetVideoFormat()) {
  }
  else {
  }

  //if (dest->GetWidth() != src->GetWidth() || dest->GetHeight() != src->GetHeight()) {
    GdkRectangle srcRect = {0, 0, src->GetWidth(), src->GetHeight()};
    GdkRectangle destRect = {0, 0,  dest->GetWidth(), dest->GetHeight()};

    dest->StretchBlit(this, &srcRect, &destRect);
  //}
  //else {
  //  memcpy(dest->GetVideoData(), src->GetVideoData(), 
  //      src->GetWidth()*src->GetHeight()*src->GetPixelSize());
 // }
} 

void FrameBuffer::DumpPMM (char *fileName) {

   FILE *file = fopen(fileName, "wb");

	fprintf(file, "P6\n");
	fprintf(file, "%d %d\n", GetWidth(), GetHeight());
	fprintf(file, "255\n");
	fwrite(GetVideoData(), 1, GetWidth()*GetHeight()*3, file);
	fclose(file);
}

/**
 */
void FrameBuffer::StretchBlit24 (FrameBuffer *src, 
			       GdkRectangle *srcRect, 
			       GdkRectangle *destRect) {

  int srcWidth = src->GetWidth();
  register unsigned char *s2Buf = ((unsigned char *)src->GetVideoData())+3*(srcRect->x+(srcRect->y*srcWidth));;
  register unsigned char *dBuf = ((unsigned char *)videoData)+3*(destRect->x+(destRect->y*width));
  register unsigned char *sBuf;

  int maxXLines = destRect->width;
  int maxYLines = destRect->height;
  int dLineWrap = 3*(width-maxXLines);
  int sLineWrap;
  register unsigned short x,y;

  if (destRect->height != srcRect->height || destRect->width != srcRect->width) {

    sLineWrap = 3*(srcWidth);

    register short yinc = ((srcRect->height / maxYLines))*sLineWrap;
    register int incyE = 2 * (srcRect->height % maxYLines);
    register int dy = incyE - 2 * maxYLines; 
    register int incyNE = dy;

    register short xinc = 3*(srcRect->width / maxXLines);
    register int incxE = 2 * (srcRect->width % maxXLines);
    register int incxNE = incxE - 2 * maxXLines;
	 register int dx;

    for (y = maxYLines; y; y--, dBuf+=dLineWrap) {
	   sBuf = s2Buf;
      dx = incxNE;
      for (x = maxXLines; x ; x--) {
	     if (dx <= 0) 
	       dx += incxE;
	     else {
	       dx += incxNE;
			 sBuf+=3;
	     }
		  sBuf+=xinc;

		  // oops, ANSI does not like this, it seems nice though... and it works
		  // *(((short *)dBuf)++) = *((short *)sBuf);
		  // well, lets do it the other way:
		  *dBuf++ = *(sBuf);
		  *dBuf++ = *(sBuf+1);
		  *dBuf++ = *(sBuf+2);
      }
      if (dy <= 0) 
	     dy += incyE;
      else {
	     dy += incyNE;
		  s2Buf+=sLineWrap;
      }
      s2Buf += yinc;
    }
  }
  else {
    sLineWrap = 3*(srcWidth-src->width);
    for (y = maxYLines; y; y--, dBuf+=dLineWrap, sBuf+=sLineWrap) {
      for (x = maxXLines; x ; x--) {
		  *dBuf++ = *sBuf++;
		  *dBuf++ = *sBuf++;
		  *dBuf++ = *sBuf++;
	   }
	 }
  }
}


/**
 * TODO macrofy me!!
 */
void FrameBuffer::StretchBlit32 (FrameBuffer *src, 
			       GdkRectangle *srcRect, 
			       GdkRectangle *destRect) {

  int srcWidth = src->GetWidth();
  register unsigned long *s2Buf = ((unsigned long *)src->GetVideoData())+srcRect->x+(srcRect->y*srcWidth);;
  register unsigned long *dBuf = ((unsigned long *)videoData)+destRect->x+(destRect->y*width);
  register unsigned long *sBuf;

  int maxXLines = destRect->width;
  int maxYLines = destRect->height;
  int dLineWrap = width-maxXLines;
  int sLineWrap;
  register unsigned short x,y;

  
  if (destRect->height != srcRect->height || destRect->width != srcRect->width) {

    sLineWrap = srcWidth;

    register short yinc = ((srcRect->height / maxYLines))*sLineWrap;
    register int incyE = 2 * (srcRect->height % maxYLines);
    register int dy = incyE - 2 * maxYLines; 
    register int incyNE = dy;

    register short xinc = srcRect->width / maxXLines;
    register int incxE = 2 * (srcRect->width % maxXLines);
    register int incxNE = incxE - 2 * maxXLines;
	 register int dx = incxNE;

    for (y = maxYLines; y; y--, dBuf+=dLineWrap) {
	   sBuf = s2Buf;
      dx = incxNE;
      for (x = maxXLines; x ; x--) {
	     if (dx <= 0) 
	       dx += incxE;
	     else {
	       dx += incxNE;
			 sBuf++;
	     }
		  sBuf+=xinc;

		  *dBuf++ = *sBuf;
      }
      if (dy <= 0) 
	     dy += incyE;
      else {
	     dy += incyNE;
		  s2Buf+=sLineWrap;
      }
      s2Buf += yinc;
    }
  }
  else {
    sLineWrap = srcWidth-src->width;
    for (register int y = maxYLines; y; y--, dBuf+=dLineWrap, sBuf+=sLineWrap) {
      for (register int x = maxXLines; x ; x--) {
		  *dBuf++ = *sBuf++;
	   }
	 }
  }
}

/**
 */
void FrameBuffer::StretchBlit (FrameBuffer *src, 
			       GdkRectangle *srcRect, 
			       GdkRectangle *destRect) {
  if (src->GetVideoFormat() == videoFormat) {
	 switch (videoFormat) {
      case RGB32:    
      case ARGB32:
		  StretchBlit32(src, srcRect, destRect);
		  return;
      case RGB24: 
		  StretchBlit24(src, srcRect, destRect);
		  return;
      default:
  	     printf ("Fast StretchBlit not supported for this videoformat\n");
		  break;
	 }
  }
  //printf ("Warning: StretchBlit will be slow for these frame buffer formats\n");

  if (destRect->height != srcRect->height || destRect->width != srcRect->width) {
    int yinc = srcRect->height / destRect->height;
    int srcys = srcRect->height % destRect->height;
    int destys = destRect->height;
    int dy = 2 * (srcys - destys); 
    int incyE = 2 * srcys;
    int incyNE = 2 * (srcys - destys);
    int sy = 0; 

    int xinc = srcRect->width / destRect->width;
    int srcxs = srcRect->width % destRect->width;
    int destxs = destRect->width;
    int incxE = 2 * srcxs;
    int incxNE = 2 * (srcxs - destxs);
	 int maxXLines = destRect->x + destRect->width;
	 int maxYLines = destRect->y + destRect->height;


    for (int y = destRect->y; y < maxYLines; y++) {
      int dx = 2 * (srcxs - destxs);
      int sx = 0;
      for (int x = destRect->x; x < maxXLines; x++) {
	if (dx <= 0) 
	  dx += incxE;
	else {
	  dx += incxNE;
	  sx ++;
	}
	sx += xinc;
	SetPixel (x, y, src->GetPixel (sx, sy));
      }
      if (dy <= 0) 
	dy += incyE;
      else {
	dy += incyNE;
	sy ++;
      }
      sy += yinc;
    }
  } else {
    for (int y = 0; y < destRect->height; y++) {
      for (int x = 0; x < destRect->width; x++) {
	SetPixel (x + destRect->x, y+ destRect->y, 
		  src->GetPixel (x + srcRect->x, y + srcRect->y));
      }
    }
  }
}



// ---------------------------------------------------------------------------
FBResolution::FBResolution () {
  name = NULL;
  width = height = 0;
  interlaced = FALSE;
  videoFormat = RGB24;
  audioFormat = None;
}

FBResolution::~FBResolution () {
  if (name)
    delete name;
}

void FBResolution::Load (XmlToken *src) {
  char *temp = src->GetString ("Name");

  if (temp)
    SetName (temp);

  width = src->GetInt ("Width");
  height = src->GetInt ("Height");
  interlaced = src->GetInt ("Interlace");
  videoFormat = (FBVideoFormat)src->GetInt ("Video");
  audioFormat = (FBAudioFormat)src->GetInt ("Audio");
}

void FBResolution::Save (XmlToken *dest) {
  if (name)
    dest->AppendString ("Name", name);
  dest->AppendInt ("Width", width);
  dest->AppendInt ("Height", height);
  dest->AppendInt ("Interlace", interlaced);
  dest->AppendInt ("Video", (int)videoFormat);
  dest->AppendInt ("Audio", (int)audioFormat);
}

void FBResolution::SetName (char *newName) {
  if (name)
    delete name;

  name = g_strdup (newName);
}

