/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __FRAMEBUFFER_H__
#define __FRAMEBUFFER_H__


#include <glib.h>
#include <assert.h>
#include <gdk/gdk.h>
#include "xml.h"

/** The format of the frame buffer video data */
enum FBVideoFormat {
  /// Standard 24-bit RGB data
  RGB24,
  /** 24-bit RGB data, packed into the low 3 bytes of 32bit words.
   * The top 8bits are undefined. */
  RGB32,
  /// 24-bit RGB data, with one byte alpha channel
  ARGB32,
  /// 16-bit YUV data
  YUV422,
  /// 16-bit YUV data planar
  YUV422P,
  /// 16-bit YUV data.
  YUV853,
  /// 16-bit RGB data
  RGB565,
  /// 15-bit RGB data (top bit is undefined)
  RGB555
};

/** The format of the frame buffer audio data.
 *  This is used by the FrameBuffer to identify the audio format used.
 */
enum FBAudioFormat {
  /// no sound 
  None,
  /// Only mono (left buffer) is used
  Mono,
  /// Normal stereo
  Stereo
};

/// Structure used to define all info about a given resolution
class FBResolution {
  /// The string name of this video format (used for selection in GUI)
  char *name;
 public:
  /// Make an empty resolution info
  FBResolution ();
  /// Destroy this resolution info
  ~FBResolution ();
  /** Set this clips name 
   *  @param newName the new name of the clip.   */
  virtual void SetName (char *newName);
  /** Get this clip's name 
   *  @return the name of the clip    */
  char *GetName () {return name;}
  /// Load a resolution.
  void Load (XmlToken *src);
  /// Save a resolution.
  void Save (XmlToken *dest);
  // The width of the video buffer
  int width;
  // The height of the frame buffer
  int height;
  /// Flag for whether buffer is interlaced
  int interlaced;
  /// The format of the video data
  FBVideoFormat videoFormat;
  /// The format of the video data
  FBAudioFormat audioFormat;
};

/** This is the main frame buffer structure. 
 *  Frame buffers are used during playback. Normally, when the program wants to display
 *  a frame, it builds a frame buffer object and asks the clip to be shown to insert
 *  itself into the frame buffer. <p>
 *  The frame buffer can have up to 3 channels. There is the image data, a left 
 *  and a right sound channel. The image data can be one of the 
 *  formats specified by the FrameBufferFormat enums and the sound channels are 
 *  always 16-bit 44.2 kHz signed.<p>
 *  to get a pointer to a specific pixel, use :<br>
 *  <tt>*pixel = GetVideoStart () + x * GetPixelSize () + y * GetStride ();</tt>, or <br>
 *  <tt>*pixel = GetLine (y) + x * GetPixelSize ();</tt><p>
 *  <p>Note: it is the job of the application to allocate the memory for the buffers 
 *        (that way, they can get direct access to in/output devices)
 */
class FrameBuffer {
  // The width of the video buffer
  int width;
  // The height of the frame buffer
  int height;
  /// The format of the video data
  FBVideoFormat videoFormat;
  /// The video data.
  unsigned char *videoData;
  /// The start of the current field (or the current frame);
  unsigned char *videoDataStart;

  /// the width (in bytes) of one scan line.
  int stride;
  /// the size (in bytes) of one pixel.
  int pixelSize;

  /// The format of the video data
  FBAudioFormat audioFormat;
  /// The audio data
  gint16 *left, *right;

  /// Flag for whether buffer is interlaced
  int interlaced;
  /// The field we are currently working on
  int currentField;
  /// Flag: Does this frame buffer have valid video data ?
  int validVideo;
  /// Flag: Does this frame buffer have valid audio data ?
  int validAudio;
 public:
  /** Make a new frame buffer.
   *  @param w the width in pixels of this frame buffer
   *  @param h the height in pixels of this frame buffer
   *  @param vidoeformat the pixel format to use
   *  @param interlaced Flag: 1 = interlaced, 0 = non-interlaced
   *  @param sound Flag: The type of sound data */
  FrameBuffer (int w, int h, FBVideoFormat videoformat = RGB24, int interlaced= 0, 
	       FBAudioFormat audioformat = None);

  /// Return the width (in pixels) of this frame buffer
  inline int GetWidth ();
  /// Return the height (in pixels) of this frame buffer
  inline int GetHeight ();

  /** Return a pointer to the start of the video data area. <br>
   *  note: This is only used by the output functions, so normally you would use
   *        the GetDataStart function instead 
   *  @return a pointer to the start of the data area */
  inline unsigned char *GetVideoData ();
  /** Set the video data pointer */
  inline void SetVideoData (unsigned char *newptr);

  /** Return a pointer to the top-left pixel's data.
   *  @note this function does not neccessarily return a pointer to the start
   *        of the data area. This is dependant on whether interlacing is active.
   *  @return a pointer to the first pixel of this frame buffer */
  inline unsigned char *GetVideoStart ();
  /** Return a pointer to the start of a given line.
   *  @param line the number of the line to get the data to
   *  @return a pointer to the first pixel of the line */
  inline unsigned char *GetLine (const int line);

  /// Return the video format.
  inline FBVideoFormat GetVideoFormat ();
  /** Get the stride of the frame buffer.
   *  @return the difference in starting address of one line and the next */
  inline int GetStride ();
  /** Get the size in bytes of one pixel */
  inline int GetPixelSize ();
  
  /** Set the given pixel to a give color.
   *  This is a function that lets you specify a color to set a pixel in ARGB 
   *  colorspace. It is <i>extremely</i> slow, so only use it for testing.
   *  @param x x-coordinate of the pixel to change
   *  @param y y-coordinate of the pixel to change
   *  @param rgba The ARGB calue of the pixel to set. the alpha value may be discarded */
  void SetPixel (int x, int y, int argb);
  /** Read a given pixel in the frame buffer.
   *  This function is <i>extremely</i> slow, so only use it for testing.
   *  @param x x-coordinate of the pixel to read
   *  @param y y-coordinate of the pixel to read
   *  @return the color of the pixel. If the buffer does not have an alpha channel
   *          this function just set alpha = 255 */
  int GetPixel (int x, int y);

  /// Get the audio format info
  inline FBAudioFormat GetAudioFormat ();
  /// Get the left audio data channel.
  inline gint16 *GetLeftAudio ();
  /// Set the left audio data pointer
  inline void SetLeftAudio (gint16 *ptr);
  /// Get the right audio data channel.
  inline gint16 *GetRightAudio ();
  /// Set the right audio data pointer
  inline void SetRightAudio (gint16 *ptr);
  /// Select the current field.
  inline void SetField (int field);
  /// Get the current field
  inline int GetField ();
  /// Get the required size of video data for this framebuffer
  inline int GetVideoDataSize ();
  /// Get info on whether this frame buffer has valid video
  inline int GetValidVideo ();
  // Set info on whether this frame buffer has valid video
  inline void SetValidVideo (int flag);
  /// Get info on whether this frame buffer has valid audio
  inline int GetValidAudio ();
  // Set info on whether this frame buffer has valid audio
  inline void SetValidAudio (int flag);
  /** Do a stretched blit from one framebuffer into another.
   *  @param src the source frame buffer.
   *  @param srcRect the rectangle in the source buffer to blit from.
   *  @param destRect the rectangle in the deestination buffer to blit to.
   */
  void StretchBlit24 (FrameBuffer *src, GdkRectangle *srcRect, GdkRectangle *destRect);
  void StretchBlit32 (FrameBuffer *src, GdkRectangle *srcRect, GdkRectangle *destRect);
  void StretchBlit (FrameBuffer *src, GdkRectangle *srcRect, GdkRectangle *destRect);

  /** converts the data in this framebuffer to the dest framebuffer, both format conversion
   * and size conversion are done.
	*/
  void ConvertInto (FrameBuffer *dest);

  void DumpPMM(char *fileName);
};

inline int FrameBuffer::GetWidth () {
  return width;
}
inline int FrameBuffer::GetHeight () {
  return height;
}

inline unsigned char *FrameBuffer::GetVideoData () {
  return videoData;
}

inline void FrameBuffer::SetVideoData (unsigned char *newptr) {
  videoData = newptr;
  SetField (currentField);
}

inline unsigned char *FrameBuffer::GetVideoStart () {
  return videoDataStart;
}
inline unsigned char *FrameBuffer::GetLine (const int line) {
  return videoDataStart + line * stride;
}

inline FBVideoFormat FrameBuffer::GetVideoFormat () {
  return videoFormat;
}
inline int FrameBuffer::GetStride () {
  return stride;
};
inline int FrameBuffer::GetPixelSize () {
  return pixelSize;
}

inline FBAudioFormat FrameBuffer::GetAudioFormat () {
  return audioFormat;
}

inline gint16 *FrameBuffer::GetLeftAudio () {
  return left;
}
inline void FrameBuffer::SetLeftAudio (gint16 *ptr) {
  left = ptr;
}
inline gint16 *FrameBuffer::GetRightAudio () {
  return right;
}
inline void FrameBuffer::SetRightAudio (gint16 *ptr) {
  right = ptr;
}
inline int FrameBuffer::GetValidAudio () {
  return validAudio;
}
inline void FrameBuffer::SetValidAudio (int flag) {
  validAudio = flag;
}
inline int FrameBuffer::GetValidVideo () {
  return validVideo;
}
inline void FrameBuffer::SetValidVideo (int flag) {
  validVideo = flag;
}


inline void FrameBuffer::SetField (int field) {
  assert (field <= interlaced && field >= 0);
  videoDataStart = videoData + field * width * pixelSize;
  currentField = field;
}

inline int FrameBuffer::GetField () {
  return currentField;
}

inline int FrameBuffer::GetVideoDataSize () {
  return height * stride;
}



#endif // __FRAMEBUFFER_H__
