/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __MCLIP_H__
#define __MCLIP_H__

#include "clip.h"
#include "framebuffer.h"
#include <gnome.h>

/** Abstract base class for all media clips.
 *  This class defines the functionality neccessary to get handle clips in
 *  timelines & to play them back. */
class MediaClip : public Clip {
  /// The edit clip top-level window.
  GtkWidget *mediaClipEdit;
  /// The widget containing the clip speed.
  GtkWidget *clipSpeed;
  /// The widget containing the clip name.
  GtkWidget *clipname;
  /// The name to revert to if 'revert' or 'cancel' is clicked.
  char *revertName;
  /// The FPS rate to revert to if 'revert' or 'cancel' is clicked.
  int revertFps;
  /// The currently active frame in the editor
  int currentFrame;

  /// The name has been changed in the edit window
  void ChangeName (GtkEntry *src);
  /// The name has been changed in the edit window
  void ChangeSpeed (GtkSpinButton *src);
  /** Create the main edit window.
   *  This function builds the edit window & fills in the 'Clip name', 'recording speed'
   *  & OK, CANCEL & REVERT buttons.
   *  @param title the title of the window.
   *  @return a GtkVbox that the edit controls can insert themselves int with 
   *  <tt>gtk_box_pack_start</tt> */
  GtkWidget *BuildEditWindow (char *title);
 protected:
  /// How many frames/sec this clip is recorded with.
  int clipFps;
  /// The IN & OUT points for this clip
  int inPoint, outPoint;
 protected:
  /** Build the class-specific bits of the edit box.
   *  The derived classes should implement this function.
   *  They should make a new notebook page and insert it into parent. */
  virtual void BuildEditWidget (GtkNotebook *parent) = NULL;
  /// 'OK' has been pressed in the edit window
  virtual void EditOkClicked ();
  /// Cancel has been pressed in the edit window 
  virtual void EditCancelClicked ();
  /// Revert has been pressed in the edit window 
  virtual void EditRevertClicked ();
  /** The edit window has been closed.
   *  if the subclasses need to do any deallocation & cleanups when the edit window gets
   *  closed, this is where they should do it. It is very important, however, that they 
   *  call CloseEditWindow in their ancestor classes */
  virtual void CloseEditWindow ();
 public:
  MediaClip ();
  virtual ~MediaClip ();

  /** Query if this clip makes use of another clip.
   *  @return 1 if [this clip] uses [clip], otherwise 0; */
  virtual int Uses (MediaClip *clip) = NULL;

  /* Render this clip's data into the current stream.
   *  @param frame the frame number of this media clip
   *  @param dest  the frame buffer to apply it to. */
  virtual void Render (int frame, int fps, FrameBuffer *dest) = NULL;

  /** Erase this clip and all associated data
   *  @return error code: 0 on success, 1 on failure */
  virtual int Erase () = NULL;

  /** Edit this clip.
   *  This function opens an edit window for the current clip. It first builds it's own 
   *  window, then calls the BuildEditWidget of the derived class. If a clip type 
   *  wishes to take complete control of the editing (e.g. timelines), this is the 
   *  function to overload. */
  virtual void Edit ();

  /** Get the length of this clip.
   *  @param  fps how many frames per second the we should convert to
   *  @return The length of the clip, measured in frames. */
  virtual int GetLength (int fps) = NULL;

  /** Get the disk size of this clip.
   *  @return The size of the clip, measured in bytes. */
  virtual int GetSize () = NULL;

  /// Set the in point of a clip.
  inline void SetInPoint (int in, int fps);
  /// Get the IN point of this clip
  inline int GetInPoint (int fps);

  /// Set the out point of a clip.
  inline void SetOutPoint (int out, int fps);
  /// Get the OUT point of this clip
  inline int GetOutPoint (int fps);

  /** Get the location of this clips file(s)
   *  @return The file or directory name used by this clip */
  virtual char *GetFileLocation () = NULL;

  /** Get cache status of this clip
   *  @return 0 if cached, 1 if not cached, 2 if not applicable */
  virtual int GetCacheStatus () = NULL;
  /// Save this clip into a token.
  virtual void Save (XmlToken *dest);
  /** Load this clip from a token.
   *  This function assumes that a clip object of the correct subclass has 
   *  already been allocated. */
  virtual void Load (XmlToken *sec);

  /** Set the current frame. This is used for editing only */
  void SetCurrentFrame (int frameNo, int fps);

  /** Get the current frame. */
  inline int GetCurrentFrame (int fps);

  /// Get how many frames/second this clip was recorded with.
  inline int GetFps ();

 private:
  /// Edit box callbacks
  static void on_revert_clicked (GtkButton *button, gpointer user_data);
  static void on_ok_clicked (GtkButton *button, gpointer user_data);
  static void on_cancel_clicked (GtkButton *button, gpointer user_data);
  static void on_name_changed (GtkEditable *editable, gpointer user_data);
  static void on_speed_changed (GtkEditable *editable, gpointer user_data);
  static gboolean on_colorEditWindow_delete_event (GtkWidget *widget, 
						   GdkEvent *event, 
						   gpointer user_data);
};

inline int MediaClip::GetInPoint (int fps) {
  return inPoint * fps / clipFps;
}
inline int MediaClip::GetOutPoint (int fps) {
  return outPoint * fps / clipFps;
}
inline void MediaClip::SetInPoint (int in, int fps) {
  inPoint = in * clipFps / fps;
}
inline void MediaClip::SetOutPoint (int out, int fps) {
  outPoint = out * clipFps / fps;
}
inline int MediaClip::GetCurrentFrame (int fps) {
  return currentFrame * fps / clipFps;
}
inline int MediaClip::GetFps () {
  return clipFps;
}


#endif
