/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __TIMELINE_H__
#define __TIMELINE_H__

#include <gnome.h>
#include "appsignal.h"
#include "timelinewindow.h"
#include "mclip.h"

/** A clip in a timeline. 
 *  This clip is a container for a clip in the timeline
 *  object. It doesn't contain the source media, but only the placement data */
class TimelineClip : public Clip {
  
  /// pointer to the source clip
  MediaClip *source;

  /** The start position of the clip. 
   *  This is measured in frames, relative to the parent. */
  int start; 

  /** The length of the clip
   *  This is measured in frames */
  int length;

  /** The IN point of the clip
   *  This marks where the entry point in the clip is. 
   *  NOTE: This is always stored in the clip's time resolution */
  int in;

  /** The layer of the clip.
   *  This is how high in the render queue that the clip is placed */
  int layer;

  /** Is this clip locked ?
   *  if TRUE, the clip cannot be dragged about in the timeline editor */
  int lock;

 public:

  /** Return a textual description of this clip type */
  virtual char *GetTypeName ();

  TimelineClip ();
  /** Get a pointer to the source clip.
   *  @return the source clip */
  MediaClip *GetSourceClip () {return source;}

  /** Set a new source clip.
   *  This function sets a new source clip, and handles all
   *  reference counts 
   *  @param newSource pointer to the new source clip   */
  void SetSourceClip (MediaClip *newSource) {source = newSource;}

  /** Set a new source clip.
   *  This function sets a new source clip, and handles all
   *  reference counts.
   *  @param newSourceName The name of the new source clip */
  void SetSourceClip (char *newSourceName);

  /** Get this clips position.
   *  @return this clips start position relative to the start of the
   *          parent, measured in frames. */
  int GetStart () {return start;}

  /** Set this clips position.
   *  @param newPos this clips position relative to the start of the
   *                parent, measured in frames. */
  void SetStart (int newStart) {start = newStart;}

  /** Get this clips layer.
   *  @return this clips layer in the render queue */
  int GetLayer () {return layer;}

  /** Set this clips layer.
   *  @param newLayer this clips layer in the render queue */
  int SetLayer (int newLayer) {layer = newLayer;}

  /** Get the IN point of the clip. */
  int GetIn () {return in;}
  /// Set the IN point of the clip
  void SetIn (int newIn) {in = newIn;}

  /** Get the ending point of the clip. */
  int GetEnd () {return start + length;}
  /// Set the ending point of the clip
  void SetEnd (int newEnd) {length = newEnd - length;}

  /// Set the length of the clip
  void SetLength (int newLength) {length = newLength;}
  /// Get the length of the clip
  int GetLength () {return length;}

  /// Add a new timeline clip (toolbox entry point).
  static void Add (GtkMenuItem *menuitem, gpointer user_data);

  /// Set the locked state of a clip.
  void SetLocked (int locked) {lock = locked;}
  /// Get the locked state of a clip.
  int GetLocked () {return lock;}
  /// Save this timeline clip to an xml token.
  virtual void Save (XmlToken *dest);
  /// Load this timeline clip from an xml token.
  virtual void Load (XmlToken *src);
};

/** A timeline object.
 *  A timeline is a clip that can contain other clips (via TimeLineClip)
 *  Each timeline can contain other clips, and remembers their position.
 *  It also implements a render queue. This works by having many layers
 *  (akin to the layers of the Gimp. Each of theese layers can contain
 *  other clips. When a frame needs to be played back, the timeline finds 
 *  All clips that have some data at the given frame. It then gets the data
 *  from theese clips, beginning at layer 0 and proceeds to layer N 
 *  @see TimelineClip. */
class Timeline : public MediaClip {
  /// Array of clips
  GSList *clips;
  /// The window used for editing
  TimelineWindow *editWindow;
 public:
  Timeline ();
  ~Timeline ();

  /** Return a textual description of this clip type */
  virtual char *GetTypeName ();

  /** Get the length of this clip.
   *  @param  fps how many frames per second the we should convert to
   *  @return The length of the clip, measured in frames. */
  virtual int GetLength (int fps);

  /** Get the disk size of this clip.
   *  @return The size of the clip, measured in bytes. */
  virtual int GetSize ();

  /** Get the location of this clips file(s)
   *  @return The file or directory name used by this clip */
  virtual char *GetFileLocation ();

  /** Get cache status of this clip
   *  @return 0 if cached, 1 if not cached, 2 if not applicable */
  virtual int GetCacheStatus ();

  /// Save this clip into a token.
  virtual void Save (XmlToken *dest);

  /// Load this Timeline from an XML token.
  virtual void Load (XmlToken *sec);

  /// Does this timeline use a given clip ?
  virtual int Uses (MediaClip *clip);
  
  /// Render a frame from this timeline into a given frame buffer
  virtual void Render (int frame, int fps, FrameBuffer *dest);

  /// Add a timeline clip to this timeline
  void AddClip (TimelineClip *clip);

  /** Add a media clip to this timeline.
   *  This function builds a timeline clip from the media clip passed to it,
   *  and inserts this timeline clip into the timeline */
  void AddClip (MediaClip *clip);

  /// Erase this clip
  virtual int Erase ();
  /** Edit this clip.
   *  This function is a stub. It's not used for timelines. */
  virtual void BuildEditWidget (GtkNotebook *parent);
  /// Create a new Timeline clip (main menu callback)
  static void Add (GtkMenuItem *menuitem, gpointer user_data);
  /** Edit entry point. 
   *  This function builds a new edit window. */
  virtual void Edit ();
  //  void BuildContents (GnomeCanvasGroup *parent, TimelineWindow *window);
  inline GSList *GetClips ();
  /// Set the frame & layer of a specific clip.
  void MoveClip (TimelineClip *clip, int newFrame, int newLayer);
  /// Set the frame & layer of a specific clip and update an editor
  void MoveClip (TimelineClip *clip, int newFrame, int newLayer,
		 TimelineEdit *editor, GnomeCanvasItem *item);
  /// Helper function to check if a specific time is free in a clip
  int IsFree (int layer, int start, int end, TimelineClip *ignore = NULL);
};

inline GSList *Timeline::GetClips () {
  return clips;
};
#endif
