#include <config.h>

#include <stdlib.h>
#include <glib.h>
#include "xml.h"
#include "misc.h"
#include <string.h>
#include <stdio.h>
#include <assert.h>

XmlToken::XmlToken () {
  key = NULL;
  contents = NULL;
  subNodes = NULL;
}

XmlToken::~XmlToken () {
  if (key)
    delete key;
  if (contents)
    delete contents;
  // TODO: Delete subnodes
}

void XmlToken::BuildText (char **text, int indent) {

  char *indstr = new char[indent + 1];
  for (int i = 0; i < indent; i++)
    indstr[i] = ' ';
  indstr[indent] = 0;
  
  if (contents) {
    char *temp = g_strdup_printf ("%s%s<%s>%s</%s>\n", *text, indstr, key, contents, key);
    delete *text;
    *text = temp;
  }

  if (subNodes) {
    GSList *t = subNodes;
    char *temp = g_strdup_printf ("%s%s<%s>\n", *text, indstr, key);
    delete *text;
    *text = temp;
    while (t) {
      ((XmlToken *)(t->data))->BuildText (text, indent + 2);
      t = t->next;
    }
    temp = g_strdup_printf ("%s%s</%s>\n", *text, indstr, key);
    delete *text;
    *text = temp;
  }
}

void XmlToken::Debug () {
  printf ("XmlToken\n--------\nKey: %s\nsubnodes: %d\ncontents: %s\n\n", key, g_slist_length (subNodes), contents);
}

int XmlToken::Load (char *filename, char *topLevel) {
  Clear ();

  FILE *inFile = fopen (filename, "rb");
  
  if (!inFile) {
    ShowDialog (filename, "Unable to open file", Error);
    return FALSE;
  }

  fseek (inFile, 0, SEEK_END);
  long len = ftell (inFile);
  fseek (inFile, 0, SEEK_SET);

  char *in = new char[len];
  fread (in, len, 1, inFile);
  fclose (inFile);
  // Find start of file
  char *temp = in;
  while (*temp == ' ' || *temp == '\n') {
    if (!*temp) {
      ShowDialog ("XML Loader", "Error: Empty file", Error);
      return FALSE;
    }
    temp++;
  }

  // Check signature
  if (strncmp (temp, "<?xml version=\"", 15)) {
    if (ShowDialog ("XML loader", "The file does not appear to have been generated by Trinity.\n\nDo you wish to attempt to load it anyways ?", YesNo) == No)
      return FALSE;
  }

  if (strncmp (&temp[15], VERSION, 5)) {
    if (ShowDialog ("XML loader", "This file was generated by another version of Trinity than you are currently using.\n\nIf you save the file, you may not be able to load it into the old version of the program", Warning) == No)
      return FALSE;
  }
  
  // Find end of signature token
  while (*temp != '>') {
    if (!*temp) {
      ShowDialog ("XML Loader", "Error: Signature token left open", Error);
      return FALSE;
    }
    temp++;
  }
  temp++;
  Parse (temp);

  char *t = NULL;
  BuildText (&t, 0);
  delete in;
  return TRUE;
}

int IsWhiteSpace (char *c) {
  switch (*c) {
  case ' ':
  case '\n':
  case '\t':
    return TRUE;
  default:
    return FALSE;
  }
}

// Find c in str
char *XmlFind (char *str, char c, int forgiving = FALSE) {
  while (*str != c) {
    switch (*str) {
    case 0:
      ShowDialog ("XML Parser", "Unexpected End-of-file", Error);
      break;
    case ' ':
    case '\n':
    case '\t':
      break;
    default: 
      if (!forgiving) {
	char *t2 = g_strndup (str - 10, 10);
	char *temp = g_strdup_printf ("Parse Error: Unexpected character: '%c'\n\nFound near '%s'", 
			      *str, t2);
	ShowDialog ("XML Parser", temp, Error);
	delete t2;
	delete temp;
	return NULL;
      }
    }
    str++;
  }
  return str;
}

char *XmlToken::Parse (char *in) {
  Clear ();

  // Skip whitespaces
  while (IsWhiteSpace (in))
    in++;

  // Find start
  char *kStart = XmlFind (in, '<');
  if (!kStart)
    return NULL;
  char *kEnd = XmlFind (kStart, '>', TRUE);
  if (!kEnd) {
    char *t2 = g_strndup (in, 20);
    char *temp = g_strdup_printf ("Error: Incomplete token start near:\n %s", t2);
    ShowDialog ("XML Parser", temp, Error);
    delete t2; 
    delete temp;
    return NULL;
  }
  kStart++;
  key = g_strndup (kStart, kEnd-kStart);

  in = kEnd + 1;

  while (IsWhiteSpace (in))
    in++;

  while (*in) {
    // Do we have a token start ?
    if (*in == '<') {
      // is it the close option
      if (in[1] == '/') {
	// Check for end key
	if (g_strncasecmp (key, &in[2], strlen (key))) {
	  char *t2 = g_strdup_printf ("Error: Token <%s> is unclosed, or\nclosing is at wrong location.", key);
	  ShowDialog ("XML parser", t2, Error);
	  delete t2;
	  return NULL;
	}
      	if (!contents && !subNodes) {
	  char *t2 = g_strdup_printf ("Error: Empty token <%s> defined.",
				      key);
	  ShowDialog ("XML parser", t2, Error);
	  delete t2;
     	  return &in[strlen (GetKey()) + 3];
	}
	Debug ();
	return &in[strlen (GetKey()) + 3];
      }
      
      if (contents) {
	char *temp = g_strdup_printf ("Error: Token <%s> has both subtoken and string contents (1)", key);

	ShowDialog ("XML parser", temp, Error);

	delete temp;
	return NULL;
      }

      XmlToken *temp = new XmlToken ();
      in = temp->Parse (in);
      if (in) 
	subNodes = g_slist_append (subNodes, temp);
      else
	return NULL;

      while (IsWhiteSpace (in))
	in++;
  
    } else {  // If not, must be a bottom-level.
      if (subNodes) {
	char *temp = g_strdup_printf ("Error: Token <%s> has both subtoken and string contents (2)", key);
	ShowDialog ("XML parser", temp, Error);
	delete temp;
	return NULL;
      }
      char *end;
      end = XmlFind (in, '<', TRUE);
      assert (!contents);
      contents = g_strndup (in, end - in);
      in = end;
    }
  }
  ShowDialog ("XML parser", "Error: Unexpected end of file.", Error);
}

void XmlToken::SetKey (char *newKey) {
  if (key)
    delete key;

  key = g_strdup (newKey);
}

void XmlToken::SetContents (char *newContents) {
  if (contents)
    delete contents;

  contents = g_strdup (newContents);
}

char *XmlToken::GetString (char *k) {
  GSList *temp = subNodes;
  while (temp) {
    if (!g_strcasecmp (((XmlToken *)(temp->data))->GetKey(), k)) {
      return ((XmlToken *)(temp->data))->GetContents();
    }
    temp = temp->next;
  }
  return NULL;
}

int XmlToken::GetInt (char *key) {
  GSList *temp = subNodes;
  while (temp) {
    if (!g_strcasecmp (((XmlToken *)(temp->data))->GetKey(),key))
      return atoi(((XmlToken *)temp->data)->GetContents());
    temp = temp->next;
  }
  return -1;
}

float XmlToken::GetFloat (char *key) {
  GSList *temp = subNodes;
  while (temp) {
    if (!g_strcasecmp (((XmlToken *)(temp->data))->GetKey(),key))
      return atof(((XmlToken *)temp->data)->GetContents());
    temp = temp->next;
  }
  return -1.0;
}

XmlToken *XmlToken::GetToken (char *key) {
  GSList *temp = subNodes;
  while (temp) {
    if (!g_strcasecmp (((XmlToken *)(temp->data))->GetKey(),key))
      return (XmlToken *)temp->data;
    temp = temp->next;
  }
  return NULL;
}

void XmlToken::AppendFloat (char *key, float val) {
  char *tempStr = g_strdup_printf ("%f", val);

  XmlToken *temp = new XmlToken ();

  temp->SetKey (key);
  temp->SetContents (tempStr);

  AppendToken (temp);
  delete tempStr;
}

void XmlToken::AppendInt (char *key, int val) {
  char *tempStr = g_strdup_printf ("%d", val);

  XmlToken *temp = new XmlToken ();

  temp->SetKey (key);
  temp->SetContents (tempStr);

  AppendToken (temp);
  delete tempStr;
}

void XmlToken::AppendString (char *key, char *val) {
  XmlToken *temp = new XmlToken ();

  temp->SetKey (key);
  temp->SetContents (val);

  AppendToken (temp);
}

void XmlToken::AppendToken (XmlToken *token) {
  subNodes = g_slist_append (subNodes, token);
}

static void DeleteSubNodes (gpointer node, gpointer dummy) {
  delete ((XmlToken *)node);
}
void XmlToken::Clear () {
  if (key)
    delete key;
  
  if (contents)
    delete contents;

  key = contents = NULL;

  if (subNodes) {
    g_slist_foreach (subNodes, DeleteSubNodes, NULL);
    g_slist_free (subNodes);
    subNodes = NULL;
  }
}

void XmlToken::Rewind () {
  scanPos = subNodes;
}

XmlToken *XmlToken::GetNext () {
  if (!scanPos)
    return NULL;
  
  XmlToken *temp = (XmlToken*)scanPos->data;
  scanPos = scanPos->next;

  return temp;
}  
