/*
 * Trinity.
 * 
 * Copyright (C) 1999 by Nicholas Francis
 * 
 * This software comes under the GPL (GNU Public License)
 * You may freely copy,distribute etc. this as long as the source code
 * is made available for FREE.
 * 
 * No warranty is made or implied. You use this program at your own risk.
 */

#ifndef __XML_H__
#define __XML_H__

#include <glib.h>

/// A Class representing a token in XML format
class XmlToken {
  /// Pointer to current next node returned by GetNext()
  GSList *scanPos;
  /// The key of this token.
  char *key;
  /// the contents of this token.
  char *contents;
  /// The list of subnodes.
  GSList *subNodes;
 public:

  XmlToken ();
  ~XmlToken ();
  /** Load a token from a file.
   *  This function opens a file, performs some simple header checking on it.
   *  If the searches for the top level token, and loads it into this.
   *  @param filename the complete path of the file to load
   *  @param toplevel the key of the top level token
   *  @return TRUE on success, FALSE on error */
  int Load (char *filename, char *topLevel);
  /** Return the contents of a token as a float.
   *  @param key the key of the token to get */
  float GetFloat (char *key);
  /** Return the contents of a token as an integer.
   *  @param key the key of the token to get */
  int GetInt (char *key);
  /** Return the contents of a token as a string.
   *  @param key the key of the token to get */
  char *GetString (char *key);
  /** Return the contents of a token as another token.
   *  @param key the key of the token to get 
   *  @return The token, or NULL if not found.*/
  XmlToken *GetToken (char *key);

  /** Print this token to stdout */
  void Debug ();

  /** Append an integer value to this token.
   *  @param key the key of the token 
   *  @param val the value of the token */
  void AppendInt (char *key, int val);
  /** Append a float value to this token.
   *  @param key the key of the token 
   *  @param val the value of the token */
  void AppendFloat (char *key, float val);
  /** Append a string value to this token.
   *  @param key the key of the token 
   *  @param val the value of the token */
  void AppendString (char *key, char *val);

  /** Append a  subtoken to this token.
   *  @param token the token to append */
  void AppendToken (XmlToken *token);

  /** Clear the contents of this token. */
  void Clear ();

  /** Write the contents of this token to a character stream.
   *  The character stream gets this token added, with a given indentation 
   *  @param dest the string to append this to 
   *  @param indent the number of spaces to indent this token's output with */
  void Write (char *dest, int indent);
  /// Set the search location to the start of the token contents
  void Rewind ();
  /** Read the next token.
   *  This function gets the next token from the input stream. Before this is called, 
   *  the <tt>Rewind</tt> function should be called.
   *  @param  dest the destination XmlToken to read the next subtoken into
   *  @return TRUE if ok, FALSE if no subtokens found */
  XmlToken *GetNext ();
  /// Set the key.
  void SetKey (char *newKey);
  /// Get the key.
  inline char *GetKey ();
  /// Set the contents.
  void SetContents (char *newContents);
  /// Get the contents string.
  char *GetContents ();
  
  /** Build an XML-compatible string that can be parsed by Load.
   *  The string should be deallocated after use. */
  void BuildText (char **text, int indent = 0);

  char *Parse (char *in);
};

inline char* XmlToken::GetKey () {
  return key;
}

inline char *XmlToken::GetContents () {
  return contents;
}

#endif
